/*
 **************************************************************************
 *
 * rz.c sz.c By Chuck Forsberg,  Omen Technology INC
 *    Copyright 1993 Omen Technology Inc All Rights Reserved
 *
 *
 *      This version implements numerous enhancements including ZMODEM
 *      Run Length Encoding and variable length headers.  These
 *      features were not funded by the original Telenet development
 *      contract.
 *
 *  This software may be freely used for educational (didactic
 *  only) purposes.  This software may also be freely used to
 *  support file transfer operations to or from licensed Omen
 *  Technology products.  Use with other commercial or shareware
 *  programs (Crosstalk, Procomm, etc.) REQUIRES REGISTRATION.
 *
 *  Any programs which use part or all of this software must be
 *  provided in source form with this notice intact except by
 *  written permission from Omen Technology Incorporated.
 *
 *
 * Use of this software for commercial or administrative purposes
 * except when exclusively limited to interfacing Omen Technology
 * products requires a per port license payment of $20.00 US per
 * port (less in quantity, see mailer.rz).  Use of this code by
 * inclusion, decompilation, reverse engineering or any other means
 * constitutes agreement to these conditions and acceptance of
 * liability to license the materials and payment of reasonable
 * legal costs necessary to enforce this license agreement.
 *
 *
 *              Omen Technology Inc
 *              Post Office Box 4681
 *              Portland OR 97208
 *
 *      This code is made available in the hope it will be useful,
 *      BUT WITHOUT ANY WARRANTY OF ANY KIND OR LIABILITY FOR ANY
 *      DAMAGES OF ANY KIND.
 *
 *  USG UNIX (3.0) ioctl conventions courtesy Jeff Martin
 *  Alarm signal handling changed to work with 4.2 BSD 7-15-84 CAF
 *
 *  SEGMENTS=n added 2-21-88 as a model for CP/M programs
 *    for CP/M-80 systems that cannot overlap modem and disk I/O.
 *
 *  compile in
 *    Directory-creating routines from Public Domain TAR by John Gilmore
 *
 *  HOWMANY may be tuned for best performance
 *  Some systems (Venix, Coherent, Regulus) may not support tty raw mode
 *  read(2) the same way as Unix. ONEREAD must be defined to force one
 *  character reads for these systems. Added 7-01-84 CAF
 *
 * Clearing for moderns POSIX systems by Vladimir Oleynik <dzo@simtreas.ru>
 * Also - set alarm minimum 30 sec for retrains modems
 */

#define VERSION "3.25 2-11-95"
#define PUBDIR "/usr/spool/uucppublic"
#define OS "POSIX"

#define NOVERBOSE

#ifndef NOVERBOSE
#define verbose_code(code) code
#else
#define verbose_code(code)
#endif


/* Hack! Special for Dionis emulation ONLY!
   NOT USE for real rz/sz!
*/
//#define DIONIS

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/time.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdarg.h>

#include <signal.h>
#include <ctype.h>
#include <errno.h>
#include <utime.h>
#include <setjmp.h>
#include <termios.h>

/*
 *   Z M O D E M . H     Manifest constants for ZMODEM
 *    application to application file transfer protocol
 */

#define ZPAD '*'        /* 052 Padding character begins frames */
#define ZDLE 030        /* Ctrl-X Zmodem escape - `ala BISYNC DLE */
#define ZDLEE (ZDLE^0100)       /* Escaped ZDLE as transmitted */
#define ZBIN 'A'        /* Binary frame indicator (CRC-16) */
#define ZHEX 'B'        /* HEX frame indicator */
#define ZBIN32 'C'      /* Binary frame with 32 bit FCS */
#define ZBINR32 'D'     /* RLE packed Binary frame with 32 bit FCS */
#define ZVBIN 'a'       /* Binary frame indicator (CRC-16) */
#define ZVHEX 'b'       /* HEX frame indicator */
#define ZVBIN32 'c'     /* Binary frame with 32 bit FCS */
#define ZVBINR32 'd'    /* RLE packed Binary frame with 32 bit FCS */
#define ZRESC   0176    /* RLE flag/escape character */
#define ZMAXHLEN 16     /* Max header information length  NEVER CHANGE */
#define ZMAXSPLEN 1024  /* Max subpacket length  NEVER CHANGE */

/* Frame types (see array "frametypes" in zm.c) */
#define ZRQINIT 0       /* Request receive init */
#define ZRINIT  1       /* Receive init */
#define ZSINIT 2        /* Send init sequence (optional) */
#define ZACK 3          /* ACK to above */
#define ZFILE 4         /* File name from sender */
#define ZSKIP 5         /* To sender: skip this file */
#define ZNAK 6          /* Last packet was garbled */
#define ZABORT 7        /* Abort batch transfers */
#define ZFIN 8          /* Finish session */
#define ZRPOS 9         /* Resume data trans at this position */
#define ZDATA 10        /* Data packet(s) follow */
#define ZEOF 11         /* End of file */
#define ZFERR 12        /* Fatal Read or Write error Detected */
#define ZCRC 13         /* Request for file CRC and response */
#define ZCHALLENGE 14   /* Receiver's Challenge */
#define ZCOMPL 15       /* Request is complete */
#define ZCAN 16         /* Other end canned session with CAN*5 */
#define ZFREECNT 17     /* Request for free bytes on filesystem */
#define ZCOMMAND 18     /* Command from sending program */
#define ZSTDERR 19      /* Output to standard error, data follows */

/* ZDLE sequences */
#define ZCRCE 'h'       /* CRC next, frame ends, header packet follows */
#define ZCRCG 'i'       /* CRC next, frame continues nonstop */
#define ZCRCQ 'j'       /* CRC next, frame continues, ZACK expected */
#define ZCRCW 'k'       /* CRC next, ZACK expected, end of frame */
#define ZRUB0 'l'       /* Translate to rubout 0177 */
#define ZRUB1 'm'       /* Translate to rubout 0377 */

/* zdlread return values (internal) */
/* -1 is general error, -2 is timeout */
#define GOTOR 0400
#define GOTCRCE (ZCRCE|GOTOR)   /* ZDLE-ZCRCE received */
#define GOTCRCG (ZCRCG|GOTOR)   /* ZDLE-ZCRCG received */
#define GOTCRCQ (ZCRCQ|GOTOR)   /* ZDLE-ZCRCQ received */
#define GOTCRCW (ZCRCW|GOTOR)   /* ZDLE-ZCRCW received */
#define GOTCAN  (GOTOR|030)     /* CAN*5 seen */

/* Byte positions within header array */
#define ZF0     3       /* First flags byte */
#define ZF1     2
#define ZF2     1
#define ZF3     0
#define ZP0     0       /* Low order 8 bits of position */
#define ZP1     1
#define ZP2     2
#define ZP3     3       /* High order 8 bits of file position */

/* Bit Masks for ZRINIT flags byte ZF0 */
#define CANFDX  01      /* Rx can send and receive true FDX */
#define CANOVIO 02      /* Rx can receive data during disk I/O */
#define CANBRK  04      /* Rx can send a break signal */
#define CANRLE  010     /* Receiver can decode RLE */
#define CANLZW  020     /* Receiver can uncompress */
#define CANFC32 040     /* Receiver can use 32 bit Frame Check */
#define ESCCTL 0100     /* Receiver expects ctl chars to be escaped */
#define ESC8   0200     /* Receiver expects 8th bit to be escaped */

/* Bit Masks for ZRINIT flags byte ZF1 */
#define CANVHDR 01      /* Variable headers OK */

/* Parameters for ZSINIT frame */
#define ZATTNLEN 32     /* Max length of attention string */
#define ALTCOFF ZF1     /* Offset to alternate canit string, 0 if not used */
/* Bit Masks for ZSINIT flags byte ZF0 */
#define TESCCTL 0100    /* Transmitter expects ctl chars to be escaped */
#define TESC8   0200    /* Transmitter expects 8th bit to be escaped */

/* Parameters for ZFILE frame */
/* Conversion options one of these in ZF0 */
#define ZCBIN   1       /* Binary transfer - inhibit conversion */
#define ZCNL    2       /* Convert NL to local end of line convention */
#define ZCRESUM 3       /* Resume interrupted file transfer */
/* Management include options, one of these ored in ZF1 */
#define ZMSKNOLOC       0200    /* Skip file if not present at rx */
/* Management options, one of these ored in ZF1 */
#define ZMMASK  037     /* Mask for the choices below */
#define ZMNEWL  1       /* Transfer if source newer or longer */
#define ZMCRC   2       /* Transfer if different file CRC or length */
#define ZMAPND  3       /* Append contents to existing file (if any) */
#define ZMCLOB  4       /* Replace existing file */
#define ZMNEW   5       /* Transfer if source newer */
	/* Number 5 is alive ... */
#define ZMDIFF  6       /* Transfer if dates or lengths different */
#define ZMPROT  7       /* Protect destination file */
#define ZMCHNG  8       /* Change filename if destination exists */
/* Transport options, one of these in ZF2 */
#define ZTLZW   1       /* Lempel-Ziv compression */
#define ZTRLE   3       /* Run Length encoding */
/* Extended options for ZF3, bit encoded */
#define ZXSPARS 64      /* Encoding for sparse file operations */
#define ZCANVHDR        01      /* Variable headers OK */
/* Receiver window size override */
#define ZRWOVR 4        /* byte position for receive window override/256 */

/* Parameters for ZCOMMAND frame ZF0 (otherwise 0) */
#define ZCACK1  1       /* Acknowledge, then do command */


static int  Twostop;             /* Use two stop bits */

/* Globals used by ZMODEM functions */
static int  Rxframeind;         /* ZBIN ZBIN32, or ZHEX type of frame */
static int  Rxtype;             /* Type of header received */
static int  Rxhlen;             /* Length of header received */
static int  Rxcount;            /* Count of data bytes received */
static char Rxhdr[ZMAXHLEN];    /* Received header */
static char Txhdr[ZMAXHLEN];    /* Transmitted header */
static int  Txfcs32;    /* TURE means send binary frames with 32 bit FCS */
static int  Crc32t;             /* Controls 32 bit CRC being sent */
				/* 1 == CRC32,  2 == CRC32 + RLE */
static int  Crc32r;     /* Indicates/controls 32 bit CRC being received */
			/* 0 == CRC16,  1 == CRC32,  2 == CRC32 + RLE */
static int  Usevhdrs;           /* Use variable length headers */
static int  Znulls;   /* Number of nulls to send at beginning of ZDATA hdr */
static char Attn[ZATTNLEN+1];   /* Attention string rx sends to tx on err */

static int lastsent;    /* Last char we sent */

static int Rxtimeout = 100;     /* Tenths of seconds to wait for something */

static unsigned long Rxpos;     /* Received file position */
static unsigned long Txpos;     /* Transmitted file position */

static int  Zmodem;              /* ZMODEM protocol requested by receiver */
static unsigned Effbaud = 115200;
static unsigned Txwindow;   /* Control the size of the transmitted window */
static unsigned Txwspac;        /* Spacing between zcrcq requests */
static unsigned Txwcnt;         /* Counter used to space ack requests */
static unsigned long Lrxpos;    /* Receiver's last reported offset */
static int  errors;
static char endmsg[128];        /* Possible message to display on exit */

static char * const frametypes[] = {
	"No Response to Error Correction Request",      /* -4 */
	"No Carrier Detect",            /* -3 */
	"TIMEOUT",              /* -2 */
	"ERROR",                /* -1 */
#define FTOFFSET 4
	"ZRQINIT",
	"ZRINIT",
	"ZSINIT",
	"ZACK",
	"ZFILE",
	"ZSKIP",
	"ZNAK",
	"ZABORT",
	"ZFIN",
	"ZRPOS",
	"ZDATA",
	"ZEOF",
	"ZFERR",
	"ZCRC",
	"ZCHALLENGE",
	"ZCOMPL",
	"ZCAN",
	"ZFREECNT",
	"ZCOMMAND",
	"ZSTDERR",
	"xxxxx"
#define FRTYPES 22      /* Total number of frame types in this array */
			/*  not including psuedo negative entries */
};


#define LOGFILE_SZ "/tmp/szlog"
#define LOGFILE2_SZ "szlog"


#define PATHLEN 257     /* ready for 4.2 bsd ? */
#define OK 0
#define FALSE 0
#ifdef TRUE
#undef TRUE
#endif
#define TRUE 1
#define ERROR (-1)

/* Ward Christensen / CP/M parameters - Don't change these! */
#define ENQ 005
#define CAN ('X'&037)
#define XOFF ('s'&037)
#define XON ('q'&037)
#define SOH 1
#define STX 2
#define EOT 4
#define ACK 6
#define NAK 025
#define CPMEOF 032
#define WANTCRC 0103    /* send C not NAK to get crc not checksum */
#define WANTG 0107      /* Send G not NAK to get nonstop batch xmsn */
#define TIMEOUT (-2)
#define RCDO (-3)
#define GCOUNT (-4)
#define RETRYMAX_SZ 10

#define ERRORMAX 5
#define RETRYMAX_RZ 5
#define WCEOT (-10)
#define UNIXFILE 0xF000 /* The S_IFMT file mask bit for stat */


/*
 * Max value for HOWMANY is 255.
 *   A larger value reduces system overhead but may evoke kernel bugs.
 */
#define HOWMANY 250

#if HOWMANY  > 255
#error Howmany must be 255 or less
#endif


#define sendline(c) net_putchar(c)

static char linbuf[HOWMANY];
static int Lleft;             /* number of characters in linbuf */
#ifdef ONEREAD
/* Sorry, Regulus and some others don't work right in raw mode! */
static int Readnum = 1; /* Number of bytes to ask for in read() from modem */
#else
static int Readnum = HOWMANY; /* Number of bytes to ask for in read() from modem */
#endif

verbose_code(static int Verbose;)

static int Filesleft;
static unsigned long Totalleft;

/*
 * Attention string to be executed by receiver to interrupt streaming data
 *  when an error is detected.  A pause (0336) may be needed before the
 *  ^C (03) or after it.
 */
static char Myattn[] = { 0 };

static FILE *in;
static FILE *fout;

static int Canseek = 1; /* 1: Can seek 0: only rewind -1: neither (pipe) */

static char txbuf[1024];
static unsigned long vpos;      /* Number of bytes read from file */

static int Lastrx;
static int Crcflg;
static int Modem2;              /* XMODEM Protocol - don't send pathnames */
static int Restricted;          /* restricted; no /.. or ../ in filenames */
static int Fullname;            /* transmit full pathname */
static int Unlinkafter;         /* Unlink file after it is sent */
static int Dottoslash;          /* Change foo.bar.baz to foo/bar/baz */
static int firstsec;
static int errcnt;              /* number of files unreadable */
static int blklen=1024;         /* length of transmitted records */
static int Optiong;             /* Let it rip no wait for sector ACK's */
static int Eofseen;             /* EOF seen on input set by zfilbuf
				   indicates cpm eof (^Z) has been received */
static int Totsecs;             /* total number of sectors this file */
static int Filcnt;              /* count of number of files opened */
static unsigned Rxbuflen=16384; /* Receiver's max buffer length */
static int Tframlen;            /* Override for tx frame length */
static int blkopt;              /* Override value for zmodem blklen */
static int Rxflags;
static unsigned long bytcnt;
static int Wantfcs32 = TRUE;    /* want to send 32 bit FCS */
static char Lzconv;             /* Local ZMODEM file conversion request */
static char Lzmanag;            /* Local ZMODEM file management request */
static int Lskipnocor;
static char Lztrans;
static int Command;             /* Send a command, then exit. */
static char *Cmdstr;            /* Pointer to the command string */
static int Cmdtries = 11;
static int Cmdack1;             /* Rx ACKs command, then do it */
static int Exitcode;
static unsigned long Lastsync;  /* Last offset to which we got a ZRPOS */
static int Beenhereb4;          /* How many times we've been ZRPOS'd here */
static int Ksendstr;            /* 1= Send esc-?-3-4-l to remote kermit */
static const char ksendbuf[] = "\033[?34l";

static jmp_buf intrjmp;         /* For the interrupt on RX CAN */
static const char *Progname;    /* the name by which we were called */


#ifdef DIONIS
static void net_flush(void)
{
	fflush(stdout);
}

static int net_rdchk(void)
{
	int waiting = 0;

	net_flush();

	if(ioctl(0, FIONREAD, &waiting)<0)
		perror("FIONREAD");
	return (!!waiting);

}

static int net_sendbrk(void)
{
	int tr = tcsendbreak(1, 0);
	if(tr<0)
		perror("tcsendbreak");
	return tr;
}

static int net_read(char *buf, int read_size)
{
	net_flush();
	return read(0, buf, read_size);
}

static void net_putchar(int c)
{
	putchar(c);
}
#else
extern int net_flush(void);
extern int net_rdchk(void);
extern int net_sendbrk(void);
extern int net_read(char *cp, int readnum);
extern int net_putchar(unsigned char c);
#endif

/*+-------------------------------------------------------------------------
	mode(new_mode)

  mode:
    0: restore original tty mode
    1: save old tty stat, set raw mode
    2: set XON/XOFF for sb/sz with ZMODEM or YMODEM-g
    3: save old tty stat, set raw mode with flow control

  if telnet, ignore the mode
--------------------------------------------------------------------------*/

#ifdef DIONIS
static struct termios oldtty, tty;

static void inittty(int new_mode)
{
	static int did0 = 0;

	switch (new_mode)
	{
		case 2:                  /* Un-raw mode used by sz,sb when -g detected */
			if (!did0)
				tcgetattr(0, &oldtty);
			tty = oldtty;
			tty.c_iflag &= ~(IXON | IXOFF | IXANY);

#if defined(RTSFLOW) && defined(CTSFLOW)
			if (tty.c_cflag & (RTSFLOW | CTSFLOW))
				tty.c_iflag = BRKINT;
			else
				tty.c_iflag = BRKINT | IXON;
#else
			tty.c_iflag = BRKINT | IXON;
#endif

			tty.c_oflag = 0; /* Transparent output */

			tty.c_cflag &= ~PARENB; /* Disable parity */
			tty.c_cflag |= CS8;     /* Set character size = 8 */
			if (Twostop)
				tty.c_cflag |= CSTOPB;  /* Set two stop bits */

			tty.c_lflag = /* Zmodem==0 ? ISIG : */ 0 ;
			/* Interrupt char */
			tty.c_cc[VINTR] = /* Zmodem==0 ? 030 : */ -1;
			tty.c_cc[VQUIT] = -1;   /* Quit char */
#if defined(NFGVMIN)
			tty.c_cc[VMIN] = 1;
#else
			tty.c_cc[VMIN] = 3;     /* This many chars satisfies reads */
#endif
			tty.c_cc[VTIME] = 1;    /* or in this many tenths of seconds */

			(void)tcsetattr(0, TCSANOW, &tty);
			did0 = 1;
			return;
		case 1:
		case 3:
			if (!did0)
				if(tcgetattr(0, &oldtty)<0) {
					perror("tcgetattr");
					exit(1);
					}
			tty = oldtty;
			tty.c_iflag &= ~(IXON | IXOFF | IXANY);

#if defined(RTSFLOW) && defined(CTSFLOW)
			tty.c_iflag = new_mode == 3 ? (IGNBRK | RTSFLOW) : IGNBRK;
#else
			tty.c_iflag = new_mode == 3 ? (IGNBRK | IXOFF) : IGNBRK;
#endif

			/* No echo,crlf mapping,INTR,QUIT,delays,no erase/kill */
			tty.c_lflag &= ~(ECHO | ICANON | ISIG);

			tty.c_oflag = 0; /* Transparent output */

			tty.c_cflag &= ~PARENB; /* Same bit rate,disable parity */
			tty.c_cflag |= CS8;     /* Set character size = 8 */
			if (Twostop)
				tty.c_cflag |= CSTOPB;  /* Set two stop bits */
#if defined(NFGVMIN)
			tty.c_cc[VMIN] = 1;     /* This many chars satisfies reads */
#else
			tty.c_cc[VMIN] = 2;    /* This many chars satisfies
											 * reads */
#endif
			tty.c_cc[VTIME] = 1;    /* or in this many tenths of seconds */
			if(tcsetattr(0, TCSANOW, &tty)<0) {
				perror("tcsetattr");
				exit(2);
			}
			did0 = 1;
			return;
		case 0:
			if (!did0)
				return;
			tcdrain(0);
			tcsetattr(0, TCSAFLUSH, &oldtty);
			return;
		default:
			return;
	}
}                                                        /* end of mode */
#else
static void inittty(int new_mode) { }
#endif

static void atexit_fun(void)
{
	inittty(0);
}

static void sendbrk(void)
{
	sleep(1);
	net_sendbrk();
	sleep(1);
	net_sendbrk();
	net_flush();
}

#ifndef NOVERBOSE
static void vfile(const char *f, ...)
{
	if (Verbose > 2) {
		va_list p;

		va_start(p, f);
		vfprintf(stderr, f, p);
		va_end(p);
		fprintf(stderr, "\n");
	}
}
#endif

/*
 * This version of readline is reasoably well suited for
 * reading many characters.
 *
 * timeout is in tenths of seconds
 */

#ifndef DIONIS
jmp_buf tohere;         /* For the interrupt on RX timeout */
/*
 * This version of readline is reasoably well suited for
 * reading many characters.
 *
 * timeout is in tenths of seconds
 */
static void alrm(int c)
{
 longjmp(tohere, -1);
}
#endif

static int readline(int timeout, int line)
{
	int n;
#ifdef DIONIS
	struct timeval tv;
	fd_set readfds;
#endif

	static char *cdq;       /* pointer for removing chars from linbuf */

	if (--Lleft >= 0) {
		verbose_code(if (Verbose > 5)
			vfile( "Calling read from line %d: Lleft=%d", line, Lleft);
		if (Verbose > 8)
			fprintf(stderr, "%02x ", *cdq&0377);)
		return (*cdq++ & 0377);
	}
	n = timeout/10;
#if 0
	if (n < 2)
		n = 2;
#else
	if (n < 30)
		n = 30;
#endif

	verbose_code(if (Verbose > 5)
	       vfile( "Calling read from line %d: alarm=%d  Readnum=%d",
		  line, n, Readnum);)

#ifdef DIONIS
	tv.tv_sec = n;
	tv.tv_usec = 0;

	FD_ZERO(&readfds);
	FD_SET(0, &readfds);
	select(1, &readfds, NULL, NULL, &tv);
	if (!FD_ISSET(0, &readfds)) {
l_tm:
		Lleft = 0;

		verbose_code(if (Verbose>1)
			vfile("Readline:TIMEOUT");)

		return TIMEOUT;
	}
#else
	if (setjmp(tohere)) {
l_tm:
		Lleft = 0;
		verbose_code(if (Verbose>1)
			fprintf(stderr, "Readline:TIMEOUT\n");)
		return TIMEOUT;
	}

	signal(SIGALRM, alrm); alarm(n);
#endif
	errno = 0;
	Lleft=net_read(cdq=linbuf, Readnum);
	if(Lleft==0) goto l_tm;
#ifndef DIONIS
	alarm(0);
#endif
	verbose_code(if (Verbose > 5)
		vfile("Read returned %d bytes errno=%d", Lleft, errno);)

	if (Lleft < 1)
		return TIMEOUT;

	verbose_code(if (Verbose > 8) {
		for (n = Lleft; --n >= 0; ) {
			fprintf(stderr, "%02x ", *cdq&0377);
		}
		vfile("");
	})
	--Lleft;
	return (*cdq++ & 0377);
}


/*
 * Purge the modem input queue of all characters
 */
static void purgeline(void)
{
	Lleft = 0;
	net_flush();
}


/*
 * Send a string to the modem, processing for \336 (sleep 1 sec)
 *   and \335 (break signal)
 */
static void zmputs(const char *s)
{
	int c;

	while (*s) {
		switch (c = *s++) {
		case '\336':
			sleep(1); continue;
		case '\335':
			sendbrk(); continue;
		default:
			sendline(c);
		}
	}
	net_flush();
}

/* send cancel string to get the other end to shut up */
static void canit(void)
{
	static const char canistr[] = {
	 24,24,24,24,24,24,24,24,24,24,8,8,8,8,8,8,8,8,8,8,0
	};

	zmputs(canistr);
	Lleft=0;        /* Do read next time ... */
}


/* called by signal interrupt or terminate to clean things up */
static void bibi_sz(int n)
{
	canit(); net_flush();
	fprintf(stderr, "%s: caught signal %d; exiting\n", Progname, n);
	if (n == SIGQUIT)
		abort();
	exit(3);
}

/* Called when ZMODEM gets an interrupt (^X) */
static void onintr(int c)
{
	signal(SIGINT, SIG_IGN);
	longjmp(intrjmp, -1);
}

static int Zctlesc;     /* Encode control characters */
static int Nozmodem;    /* If invoked as "sb" or "rb"*/

static int Zrwindow = 1400;     /* RX window size (controls garbage count) */

/*
 * Log an error
 */

#ifndef NOVERBOSE
static void zperr(const char *s, ...)
{
	va_list p;

	if (Verbose <= 0)
		return;
	fprintf(stderr, "Retry %d: ", errors);
	va_start(p, s);
	vfprintf(stderr, s, p);
	va_end(p);
	fprintf(stderr, "\n");
}
#endif

/*
 * If called as sb use YMODEM protocol
 */
static void chkinvok_sz(const char *s)
{
	if (s[0]=='z' && s[1] == 'c') {
		Command = TRUE;
		if (s[8] == 'i')
			Cmdack1 = ZCACK1;
	}
	if (s[0]=='s' && s[1]=='b') {
		Nozmodem = TRUE; blklen=1024;
	}
	if (s[0]=='s' && s[1]=='x') {
		Modem2 = TRUE;
	}
}

static void usage_sz(void)
{
	inittty(0);
	fprintf(stderr,
"\n%s %s for %s by Chuck Forsberg, Omen Technology INC\n"
"\tported to telnet by Heo, Junhyeok, KAIST\n"
"\t\t\042The High Reliability Software\042\n"
"Send Files and Commands with ZMODEM/YMODEM/XMODEM Protocol\n\n"
"Usage: sz [-2+abcdefgklLnNuvwyY] [-] file ...\n"
"\t     zcommand [-2Cegv] COMMAND\n"
"\t     zcommandi [-2Cegv] COMMAND\n"
"\t     sb [-2adfkuv] [-] file ...\n"
"\t     sx [-2akuv] [-] file\n"
"\nCopyright 1993 Omen Technology INC All Rights Reserved\n"
"See sz.doc for option descriptions and licensing information.\n\n"
"Ported into telnet by Heo, Junhyeok, 1994.3.28\n"
"Linux porting by Park Myeong Seok,   1995.4.23\n"
	, Progname, VERSION, OS);
	exit(3);
}

/*
 *  Crc calculation stuff
 */

/* crctab calculated by Mark G. Mendel, Network Systems Corporation */
static const unsigned short crctab[256] = {
    0x0000,  0x1021,  0x2042,  0x3063,  0x4084,  0x50a5,  0x60c6,  0x70e7,
    0x8108,  0x9129,  0xa14a,  0xb16b,  0xc18c,  0xd1ad,  0xe1ce,  0xf1ef,
    0x1231,  0x0210,  0x3273,  0x2252,  0x52b5,  0x4294,  0x72f7,  0x62d6,
    0x9339,  0x8318,  0xb37b,  0xa35a,  0xd3bd,  0xc39c,  0xf3ff,  0xe3de,
    0x2462,  0x3443,  0x0420,  0x1401,  0x64e6,  0x74c7,  0x44a4,  0x5485,
    0xa56a,  0xb54b,  0x8528,  0x9509,  0xe5ee,  0xf5cf,  0xc5ac,  0xd58d,
    0x3653,  0x2672,  0x1611,  0x0630,  0x76d7,  0x66f6,  0x5695,  0x46b4,
    0xb75b,  0xa77a,  0x9719,  0x8738,  0xf7df,  0xe7fe,  0xd79d,  0xc7bc,
    0x48c4,  0x58e5,  0x6886,  0x78a7,  0x0840,  0x1861,  0x2802,  0x3823,
    0xc9cc,  0xd9ed,  0xe98e,  0xf9af,  0x8948,  0x9969,  0xa90a,  0xb92b,
    0x5af5,  0x4ad4,  0x7ab7,  0x6a96,  0x1a71,  0x0a50,  0x3a33,  0x2a12,
    0xdbfd,  0xcbdc,  0xfbbf,  0xeb9e,  0x9b79,  0x8b58,  0xbb3b,  0xab1a,
    0x6ca6,  0x7c87,  0x4ce4,  0x5cc5,  0x2c22,  0x3c03,  0x0c60,  0x1c41,
    0xedae,  0xfd8f,  0xcdec,  0xddcd,  0xad2a,  0xbd0b,  0x8d68,  0x9d49,
    0x7e97,  0x6eb6,  0x5ed5,  0x4ef4,  0x3e13,  0x2e32,  0x1e51,  0x0e70,
    0xff9f,  0xefbe,  0xdfdd,  0xcffc,  0xbf1b,  0xaf3a,  0x9f59,  0x8f78,
    0x9188,  0x81a9,  0xb1ca,  0xa1eb,  0xd10c,  0xc12d,  0xf14e,  0xe16f,
    0x1080,  0x00a1,  0x30c2,  0x20e3,  0x5004,  0x4025,  0x7046,  0x6067,
    0x83b9,  0x9398,  0xa3fb,  0xb3da,  0xc33d,  0xd31c,  0xe37f,  0xf35e,
    0x02b1,  0x1290,  0x22f3,  0x32d2,  0x4235,  0x5214,  0x6277,  0x7256,
    0xb5ea,  0xa5cb,  0x95a8,  0x8589,  0xf56e,  0xe54f,  0xd52c,  0xc50d,
    0x34e2,  0x24c3,  0x14a0,  0x0481,  0x7466,  0x6447,  0x5424,  0x4405,
    0xa7db,  0xb7fa,  0x8799,  0x97b8,  0xe75f,  0xf77e,  0xc71d,  0xd73c,
    0x26d3,  0x36f2,  0x0691,  0x16b0,  0x6657,  0x7676,  0x4615,  0x5634,
    0xd94c,  0xc96d,  0xf90e,  0xe92f,  0x99c8,  0x89e9,  0xb98a,  0xa9ab,
    0x5844,  0x4865,  0x7806,  0x6827,  0x18c0,  0x08e1,  0x3882,  0x28a3,
    0xcb7d,  0xdb5c,  0xeb3f,  0xfb1e,  0x8bf9,  0x9bd8,  0xabbb,  0xbb9a,
    0x4a75,  0x5a54,  0x6a37,  0x7a16,  0x0af1,  0x1ad0,  0x2ab3,  0x3a92,
    0xfd2e,  0xed0f,  0xdd6c,  0xcd4d,  0xbdaa,  0xad8b,  0x9de8,  0x8dc9,
    0x7c26,  0x6c07,  0x5c64,  0x4c45,  0x3ca2,  0x2c83,  0x1ce0,  0x0cc1,
    0xef1f,  0xff3e,  0xcf5d,  0xdf7c,  0xaf9b,  0xbfba,  0x8fd9,  0x9ff8,
    0x6e17,  0x7e36,  0x4e55,  0x5e74,  0x2e93,  0x3eb2,  0x0ed1,  0x1ef0
};

/*
 * updcrc macro derived from article Copyright (C) 1986 Stephen Satchell.
 *  NOTE: First srgument must be in range 0 to 255.
 *        Second argument is referenced twice.
 *
 * Programmers may incorporate any or all code into their programs,
 * giving proper credit within the source. Publication of the
 * source routines is permitted so long as proper credit is given
 * to Stephen Satchell, Satchell Evaluations and Chuck Forsberg,
 * Omen Technology.
 */

#define updcrc(cp, crc) ( crctab[((crc >> 8) & 255)] ^ (crc << 8) ^ cp)

/*
 * Copyright (C) 1986 Gary S. Brown.  You may use this program, or
 * code or tables extracted from it, as desired without restriction.
 */

/* First, the polynomial itself and its table of feedback terms.  The  */
/* polynomial is                                                       */
/* X^32+X^26+X^23+X^22+X^16+X^12+X^11+X^10+X^8+X^7+X^5+X^4+X^2+X^1+X^0 */
/* Note that we take it "backwards" and put the highest-order term in  */
/* the lowest-order bit.  The X^32 term is "implied"; the LSB is the   */
/* X^31 term, etc.  The X^0 term (usually shown as "+1") results in    */
/* the MSB being 1.                                                    */

/* Note that the usual hardware shift register implementation, which   */
/* is what we're using (we're merely optimizing it by doing eight-bit  */
/* chunks at a time) shifts bits into the lowest-order term.  In our   */
/* implementation, that means shifting towards the right.  Why do we   */
/* do it this way?  Because the calculated CRC must be transmitted in  */
/* order from highest-order term to lowest-order term.  UARTs transmit */
/* characters in order from LSB to MSB.  By storing the CRC this way,  */
/* we hand it to the UART in the order low-byte to high-byte; the UART */
/* sends each low-bit to hight-bit; and the result is transmission bit */
/* by bit from highest- to lowest-order term without requiring any bit */
/* shuffling on our part.  Reception works similarly.                  */

/* The feedback terms table consists of 256, 32-bit entries.  Notes:   */
/*                                                                     */
/*     The table can be generated at runtime if desired; code to do so */
/*     is shown later.  It might not be obvious, but the feedback      */
/*     terms simply represent the results of eight shift/xor opera-    */
/*     tions for all combinations of data and CRC register values.     */
/*                                                                     */
/*     The values must be right-shifted by eight bits by the "updcrc"  */
/*     logic; the shift must be unsigned (bring in zeroes).  On some   */
/*     hardware you could probably optimize the shift in assembler by  */
/*     using byte-swap instructions.                                   */

static const long cr3tab[] = { /* CRC polynomial 0xedb88320 */
0x00000000, 0x77073096, 0xee0e612c, 0x990951ba, 0x076dc419, 0x706af48f, 0xe963a535, 0x9e6495a3,
0x0edb8832, 0x79dcb8a4, 0xe0d5e91e, 0x97d2d988, 0x09b64c2b, 0x7eb17cbd, 0xe7b82d07, 0x90bf1d91,
0x1db71064, 0x6ab020f2, 0xf3b97148, 0x84be41de, 0x1adad47d, 0x6ddde4eb, 0xf4d4b551, 0x83d385c7,
0x136c9856, 0x646ba8c0, 0xfd62f97a, 0x8a65c9ec, 0x14015c4f, 0x63066cd9, 0xfa0f3d63, 0x8d080df5,
0x3b6e20c8, 0x4c69105e, 0xd56041e4, 0xa2677172, 0x3c03e4d1, 0x4b04d447, 0xd20d85fd, 0xa50ab56b,
0x35b5a8fa, 0x42b2986c, 0xdbbbc9d6, 0xacbcf940, 0x32d86ce3, 0x45df5c75, 0xdcd60dcf, 0xabd13d59,
0x26d930ac, 0x51de003a, 0xc8d75180, 0xbfd06116, 0x21b4f4b5, 0x56b3c423, 0xcfba9599, 0xb8bda50f,
0x2802b89e, 0x5f058808, 0xc60cd9b2, 0xb10be924, 0x2f6f7c87, 0x58684c11, 0xc1611dab, 0xb6662d3d,
0x76dc4190, 0x01db7106, 0x98d220bc, 0xefd5102a, 0x71b18589, 0x06b6b51f, 0x9fbfe4a5, 0xe8b8d433,
0x7807c9a2, 0x0f00f934, 0x9609a88e, 0xe10e9818, 0x7f6a0dbb, 0x086d3d2d, 0x91646c97, 0xe6635c01,
0x6b6b51f4, 0x1c6c6162, 0x856530d8, 0xf262004e, 0x6c0695ed, 0x1b01a57b, 0x8208f4c1, 0xf50fc457,
0x65b0d9c6, 0x12b7e950, 0x8bbeb8ea, 0xfcb9887c, 0x62dd1ddf, 0x15da2d49, 0x8cd37cf3, 0xfbd44c65,
0x4db26158, 0x3ab551ce, 0xa3bc0074, 0xd4bb30e2, 0x4adfa541, 0x3dd895d7, 0xa4d1c46d, 0xd3d6f4fb,
0x4369e96a, 0x346ed9fc, 0xad678846, 0xda60b8d0, 0x44042d73, 0x33031de5, 0xaa0a4c5f, 0xdd0d7cc9,
0x5005713c, 0x270241aa, 0xbe0b1010, 0xc90c2086, 0x5768b525, 0x206f85b3, 0xb966d409, 0xce61e49f,
0x5edef90e, 0x29d9c998, 0xb0d09822, 0xc7d7a8b4, 0x59b33d17, 0x2eb40d81, 0xb7bd5c3b, 0xc0ba6cad,
0xedb88320, 0x9abfb3b6, 0x03b6e20c, 0x74b1d29a, 0xead54739, 0x9dd277af, 0x04db2615, 0x73dc1683,
0xe3630b12, 0x94643b84, 0x0d6d6a3e, 0x7a6a5aa8, 0xe40ecf0b, 0x9309ff9d, 0x0a00ae27, 0x7d079eb1,
0xf00f9344, 0x8708a3d2, 0x1e01f268, 0x6906c2fe, 0xf762575d, 0x806567cb, 0x196c3671, 0x6e6b06e7,
0xfed41b76, 0x89d32be0, 0x10da7a5a, 0x67dd4acc, 0xf9b9df6f, 0x8ebeeff9, 0x17b7be43, 0x60b08ed5,
0xd6d6a3e8, 0xa1d1937e, 0x38d8c2c4, 0x4fdff252, 0xd1bb67f1, 0xa6bc5767, 0x3fb506dd, 0x48b2364b,
0xd80d2bda, 0xaf0a1b4c, 0x36034af6, 0x41047a60, 0xdf60efc3, 0xa867df55, 0x316e8eef, 0x4669be79,
0xcb61b38c, 0xbc66831a, 0x256fd2a0, 0x5268e236, 0xcc0c7795, 0xbb0b4703, 0x220216b9, 0x5505262f,
0xc5ba3bbe, 0xb2bd0b28, 0x2bb45a92, 0x5cb36a04, 0xc2d7ffa7, 0xb5d0cf31, 0x2cd99e8b, 0x5bdeae1d,
0x9b64c2b0, 0xec63f226, 0x756aa39c, 0x026d930a, 0x9c0906a9, 0xeb0e363f, 0x72076785, 0x05005713,
0x95bf4a82, 0xe2b87a14, 0x7bb12bae, 0x0cb61b38, 0x92d28e9b, 0xe5d5be0d, 0x7cdcefb7, 0x0bdbdf21,
0x86d3d2d4, 0xf1d4e242, 0x68ddb3f8, 0x1fda836e, 0x81be16cd, 0xf6b9265b, 0x6fb077e1, 0x18b74777,
0x88085ae6, 0xff0f6a70, 0x66063bca, 0x11010b5c, 0x8f659eff, 0xf862ae69, 0x616bffd3, 0x166ccf45,
0xa00ae278, 0xd70dd2ee, 0x4e048354, 0x3903b3c2, 0xa7672661, 0xd06016f7, 0x4969474d, 0x3e6e77db,
0xaed16a4a, 0xd9d65adc, 0x40df0b66, 0x37d83bf0, 0xa9bcae53, 0xdebb9ec5, 0x47b2cf7f, 0x30b5ffe9,
0xbdbdf21c, 0xcabac28a, 0x53b39330, 0x24b4a3a6, 0xbad03605, 0xcdd70693, 0x54de5729, 0x23d967bf,
0xb3667a2e, 0xc4614ab8, 0x5d681b02, 0x2a6f2b94, 0xb40bbe37, 0xc30c8ea1, 0x5a05df1b, 0x2d02ef8d
};

#define UPDC32(b, c) (cr3tab[((int)c ^ b) & 0xff] ^ ((c >> 8) & 0x00FFFFFF))

#ifndef NOVERBOSE
static const char badcrc[] = "Bad CRC";
static const char * const Zendnames[] = { "ZCRCE", "ZCRCG", "ZCRCQ", "ZCRCW"};
#endif

/*
 * Send character c with ZMODEM escape sequence encoding.
 *  Escape XON, XOFF. Escape CR following @ (Telenet net escape)
 */
static void zsendline(int c)
{

	/* Quick check for non control characters */
	if (c & 0140)
		net_putchar(lastsent = c);
	else {
		switch (c &= 0377) {
		case ZDLE:
			net_putchar(ZDLE);
			net_putchar (lastsent = (c ^= 0100));
			break;
		case 015:
		case 0215:
			if (!Zctlesc && (lastsent & 0177) != '@')
				goto sendit;
		/* **** FALL THRU TO **** */
		case 020:
		case 021:
		case 023:
		case 0220:
		case 0221:
		case 0223:
			net_putchar(ZDLE);
			c ^= 0100;
	sendit:
			net_putchar(lastsent = c);
			break;
		default:
			if (Zctlesc && ! (c & 0140)) {
				net_putchar(ZDLE);
				c ^= 0100;
			}
			net_putchar(lastsent = c);
		}
	}
}

/* Send data subpacket RLE encoded with 32 bit FCS */
static void zsdar32(const char * buf, int length, int frameend)
{
	int c, l, n;
	unsigned long crc;

	crc = 0xFFFFFFFFL;  l = *buf++ & 0377;
	if (length == 1) {
		zsendline(l); crc = UPDC32(l, crc);
		if (l == ZRESC) {
			zsendline(1); crc = UPDC32(1, crc);
		}
	} else {
		for (n = 0; --length >= 0; ++buf) {
			if ((c = *buf & 0377) == l && n < 126 && length>0) {
				++n;  continue;
			}
			switch (n) {
			case 0:
				zsendline(l);
				crc = UPDC32(l, crc);
				if (l == ZRESC) {
					zsendline(0100); crc = UPDC32(0100, crc);
				}
				l = c; break;
			case 1:
				if (l != ZRESC) {
					zsendline(l); zsendline(l);
					crc = UPDC32(l, crc);
					crc = UPDC32(l, crc);
					n = 0; l = c; break;
				}
				/* **** FALL THRU TO **** */
			default:
				zsendline(ZRESC); crc = UPDC32(ZRESC, crc);
				if (l == 040 && n < 34) {
					n += 036;
					zsendline(n); crc = UPDC32(n, crc);
				}
				else {
					n += 0101;
					zsendline(n); crc = UPDC32(n, crc);
					zsendline(l); crc = UPDC32(l, crc);
				}
				n = 0; l = c; break;
			}
		}
	}
	net_putchar(ZDLE); net_putchar(frameend);
	crc = UPDC32(frameend, crc);

	crc = ~crc;
	for (length=4; --length >= 0;) {
		zsendline((int)crc);  crc >>= 8;
	}
}

/*
 * Read a byte, checking for ZMODEM escape encoding
 *  including CAN*5 which represents a quick abort
 */
static int zdlread(void)
{
	int c;

again:
	/* Quick check for non control characters */
	if ((c = readline(Rxtimeout, __LINE__)) & 0140)
		return c;
	switch (c) {
	case ZDLE:
		break;
	case 023:
	case 0223:
	case 021:
	case 0221:
		goto again;
	default:
		if (Zctlesc && !(c & 0140)) {
			goto again;
		}
		return c;
	}
again2:
	if ((c = readline(Rxtimeout, __LINE__)) < 0)
		return c;
	if (c == CAN && (c = readline(Rxtimeout, __LINE__)) < 0)
		return c;
	if (c == CAN && (c = readline(Rxtimeout, __LINE__)) < 0)
		return c;
	if (c == CAN && (c = readline(Rxtimeout, __LINE__)) < 0)
		return c;
	switch (c) {
	case CAN:
		return GOTCAN;
	case ZCRCE:
	case ZCRCG:
	case ZCRCQ:
	case ZCRCW:
		return (c | GOTOR);
	case ZRUB0:
		return 0177;
	case ZRUB1:
		return 0377;
	case 023:
	case 0223:
	case 021:
	case 0221:
		goto again2;
	default:
		if (Zctlesc && ! (c & 0140)) {
			goto again2;
		}
		if ((c & 0140) ==  0100)
			return (c ^ 0100);
		break;
	}
	verbose_code(if (Verbose>1)
		zperr("Bad escape sequence %x", c);)
	return ERROR;
}


/* Receive data subpacket RLE encoded with 32 bit FCS */
static int zrdatr32(char *buf, int length)
{
	int c;
	unsigned long crc;
	char *end;
	int d;

	crc = 0xFFFFFFFFL;  Rxcount = 0;  end = buf + length;
	d = 0;  /* Use for RLE decoder state */
	while (buf <= end) {
		if ((c = zdlread()) & ~0377) {
crcfoo:
			switch (c) {
			case GOTCRCE:
			case GOTCRCG:
			case GOTCRCQ:
			case GOTCRCW:
				d = c;  c &= 0377;
				crc = UPDC32(c, crc);
				if ((c = zdlread()) & ~0377)
					goto crcfoo;
				crc = UPDC32(c, crc);
				if ((c = zdlread()) & ~0377)
					goto crcfoo;
				crc = UPDC32(c, crc);
				if ((c = zdlread()) & ~0377)
					goto crcfoo;
				crc = UPDC32(c, crc);
				if ((c = zdlread()) & ~0377)
					goto crcfoo;
				crc = UPDC32(c, crc);
				if (crc != 0xDEBB20E3) {
					verbose_code(zperr(badcrc);)
					return ERROR;
				}
				Rxcount = length - (end - buf);
				verbose_code(vfile("zrdatr32: %d %s", Rxcount,
				  Zendnames[(d-GOTCRCE)&3]);)
				return d;
			case GOTCAN:
				verbose_code(zperr("Sender Canceled");)
				return ZCAN;
			case TIMEOUT:
				verbose_code(zperr("TIMEOUT");)
				return c;
			default:
				verbose_code(zperr("Bad data subpacket");)
				return c;
			}
		}
		crc = UPDC32(c, crc);
		switch (d) {
		case 0:
			if (c == ZRESC) {
				d = -1;  continue;
			}
			*buf++ = c;  continue;
		case -1:
			if (c >= 040 && c < 0100) {
				d = c - 035; c = 040;  goto spaces;
			}
			if (c == 0100) {
				d = 0;
				*buf++ = ZRESC;  continue;
			}
			d = c;  continue;
		default:
			d -= 0100;
			if (d < 1)
				goto badpkt;
spaces:
			if ((buf + d) > end)
				goto badpkt;
			while ( --d >= 0)
				*buf++ = c;
			d = 0;  continue;
		}
	}
badpkt:
	verbose_code(zperr("Data subpacket too long");)
	return ERROR;
}

/*
 * Entry point Functions:
 *      zsbhdr(type, hdr) send binary header
 *      zshhdr(type, hdr) send hex header
 *      zgethdr(hdr, eflag) receive header - binary or hex
 *      zsdata(buf, len, frameend) send data
 *      zrdata(buf, len) receive data
 *      stohdr(pos) store position data in Txhdr
 *      unsigned long rclhdr(hdr) recover position offset from header
 *
 *
 */


/* Recover a unsigned long integer from a header */
static unsigned long rclhdr(const char *hdr)
{
	unsigned long l;

	l = (hdr[ZP3] & 0377);
	l = (l << 8) | (hdr[ZP2] & 0377);
	l = (l << 8) | (hdr[ZP1] & 0377);
	l = (l << 8) | (hdr[ZP0] & 0377);
	return l;
}

/* Send ZMODEM binary header hdr of type type */
static void zsbh32(int len, const char *hdr, int type, int flavour)
{
	int n;
	unsigned long crc;

	net_putchar(flavour);
	if (Usevhdrs)
		zsendline(len);
	zsendline(type);
	crc = 0xFFFFFFFFL; crc = UPDC32(type, crc);

	for (n=len; --n >= 0; ++hdr) {
		crc = UPDC32((0377 & *hdr), crc);
		zsendline(*hdr);
	}
	crc = ~crc;
	for (n=4; --n >= 0;) {
		zsendline((int)crc);
		crc >>= 8;
	}
}


/* Send ZMODEM binary header hdr of type type */
static void zsbhdr(int len, int type, const char *hdr)
{
	int n;
	unsigned short crc;

	verbose_code(vfile("zsbhdr: %c %d %s %lx", Usevhdrs?'v':'f', len,
	  frametypes[type+FTOFFSET], rclhdr(hdr));)
	if (type == ZDATA)
		for (n = Znulls; --n >=0; )
			net_putchar(0);

	net_putchar(ZPAD); net_putchar(ZDLE);

	switch (Crc32t=Txfcs32) {
	case 2:
		zsbh32(len, hdr, type, Usevhdrs?ZVBINR32:ZBINR32);
		net_flush();  break;
	case 1:
		zsbh32(len, hdr, type, Usevhdrs?ZVBIN32:ZBIN32);  break;
	default:
		if (Usevhdrs) {
			net_putchar(ZVBIN);
			zsendline(len);
		}
		else
			net_putchar(ZBIN);
		zsendline(type);
		crc = updcrc(type, 0);

		for (n=len; --n >= 0; ++hdr) {
			zsendline(*hdr);
			crc = updcrc((0377& *hdr), crc);
		}
		crc = updcrc(0,updcrc(0,crc));
		zsendline(crc>>8);
		zsendline(crc);
	}
	if (type != ZDATA)
		net_flush();
}

/* Send a byte as two hex digits */
static void zputhex(int c)
{
	static char     digits[]        = "0123456789abcdef";

	verbose_code(if (Verbose>8)
		vfile("zputhex: %02X", c);)
	sendline(digits[(c&0xF0)>>4]);
	sendline(digits[(c)&0xF]);
}

/* Send ZMODEM HEX header hdr of type type */
static void zshhdr(int len, int type, const char *hdr)
{
	int n;
	unsigned short crc;

	verbose_code(vfile("zshhdr: %c %d %s %lx", Usevhdrs?'v':'f', len,
	  frametypes[type+FTOFFSET], rclhdr(hdr));)
	sendline(ZPAD); sendline(ZPAD); sendline(ZDLE);
	if (Usevhdrs) {
		sendline(ZVHEX);
		zputhex(len);
	}
	else
		sendline(ZHEX);
	zputhex(type);
	Crc32t = 0;

	crc = updcrc(type, 0);
	for (n=len; --n >= 0; ++hdr) {
		zputhex(*hdr); crc = updcrc((0377 & *hdr), crc);
	}
	crc = updcrc(0,updcrc(0,crc));
	zputhex(crc>>8); zputhex(crc);

	/* Make it printable on remote machine */
	sendline(015); sendline(0212);
	/*
	 * Uncork the remote in case a fake XOFF has stopped data flow
	 */
	if (type != ZFIN && type != ZACK)
		sendline(021);
	net_flush();
}


static void zsda32(const char *buf, int length, int frameend)
{
	int c;
	unsigned long crc;

	crc = 0xFFFFFFFFL;
	for (;--length >= 0; ++buf) {
		c = *buf & 0377;
		if (c & 0140)
			net_putchar(lastsent = c);
		else
			zsendline(c);
		crc = UPDC32(c, crc);
	}
	net_putchar(ZDLE); net_putchar(frameend);
	crc = UPDC32(frameend, crc);

	crc = ~crc;
	for (c=4; --c >= 0;) {
		zsendline((int)crc);  crc >>= 8;
	}
}


/*
 * Send binary array buf of length length, with ending ZDLE sequence frameend
 */

static void zsdata(const char *buf, int length, int frameend)
{
	unsigned short crc;

	verbose_code(vfile("zsdata: %d %s", length, Zendnames[(frameend-ZCRCE)&3]);)
	switch (Crc32t) {
	case 1:
		zsda32(buf, length, frameend);  break;
	case 2:
		zsdar32(buf, length, frameend);  break;
	default:
		crc = 0;
		for (;--length >= 0; ++buf) {
			zsendline(*buf); crc = updcrc((0377 & *buf), crc);
		}
		net_putchar(ZDLE); net_putchar(frameend);
		crc = updcrc(frameend, crc);

		crc = updcrc(0,updcrc(0,crc));
		zsendline(crc>>8); zsendline(crc);
	}
	if (frameend == ZCRCW)
		net_putchar(XON);
	if (frameend != ZCRCG)
		net_flush();
}


#ifndef NOVERBOSE
static void garbitch(void)
{
	zperr("Garbled data subpacket");
}
#endif

static int zrdat32(char *buf, int length)
{
	int c;
	unsigned long crc;
	char *end;
	int d;

	crc = 0xFFFFFFFFL;  Rxcount = 0;  end = buf + length;
	while (buf <= end) {
		if ((c = zdlread()) & ~0377) {
crcfoo:
			switch (c) {
			case GOTCRCE:
			case GOTCRCG:
			case GOTCRCQ:
			case GOTCRCW:
				d = c;  c &= 0377;
				crc = UPDC32(c, crc);
				if ((c = zdlread()) & ~0377)
					goto crcfoo;
				crc = UPDC32(c, crc);
				if ((c = zdlread()) & ~0377)
					goto crcfoo;
				crc = UPDC32(c, crc);
				if ((c = zdlread()) & ~0377)
					goto crcfoo;
				crc = UPDC32(c, crc);
				if ((c = zdlread()) & ~0377)
					goto crcfoo;
				crc = UPDC32(c, crc);
				if (crc != 0xDEBB20E3) {
					verbose_code(zperr(badcrc);)
					return ERROR;
				}
				Rxcount = length - (end - buf);
				verbose_code(vfile("zrdat32: %d %s", Rxcount,
				 Zendnames[(d-GOTCRCE)&3]);)
				return d;
			case GOTCAN:
				verbose_code(zperr("Sender Canceled");)
				return ZCAN;
			case TIMEOUT:
				verbose_code(zperr("TIMEOUT");)
				return c;
			default:
				verbose_code(garbitch();)
				return c;
			}
		}
		*buf++ = c;
		crc = UPDC32(c, crc);
	}
	verbose_code(zperr("Data subpacket too long");)
	return ERROR;
}


/*
 * Receive array buf of max length with ending ZDLE sequence
 *  and CRC.  Returns the ending character or error code.
 *  NB: On errors may store length+1 bytes!
 */
static int zrdata(char *buf, int length)
{
	int c;
	unsigned short crc;
	char *end;
	int d;

	switch (Crc32r) {
	case 1:
		return zrdat32(buf, length);
	case 2:
		return zrdatr32(buf, length);
	}

	crc = Rxcount = 0;  end = buf + length;
	while (buf <= end) {
		if ((c = zdlread()) & ~0377) {
crcfoo:
			switch (c) {
			case GOTCRCE:
			case GOTCRCG:
			case GOTCRCQ:
			case GOTCRCW:
				d = c;
				crc = updcrc((c&0377), crc);
				if ((c = zdlread()) & ~0377)
					goto crcfoo;
				crc = updcrc(c, crc);
				if ((c = zdlread()) & ~0377)
					goto crcfoo;
				crc = updcrc(c, crc);
				if (crc & 0xFFFF) {
					verbose_code(zperr(badcrc);)
					return ERROR;
				}
				Rxcount = length - (end - buf);
				verbose_code(vfile("zrdata: %d  %s", Rxcount,
				 Zendnames[(d-GOTCRCE)&3]);)
				return d;
			case GOTCAN:
				verbose_code(zperr("Sender Canceled");)
				return ZCAN;
			case TIMEOUT:
				verbose_code(zperr("TIMEOUT");)
				return c;
			default:
				verbose_code(garbitch();)
				return c;
			}
		}
		*buf++ = c;
		crc = updcrc(c, crc);
	}
	verbose_code(zperr("Data subpacket too long");)
	return ERROR;
}


/* Receive a binary style header (type and position) */
static int zrbhdr(char *hdr)
{
	int c, n;
	unsigned short crc;

	if ((c = zdlread()) & ~0377)
		return c;
	Rxtype = c;
	crc = updcrc(c, 0);

	for (n=Rxhlen; --n >= 0; ++hdr) {
		if ((c = zdlread()) & ~0377)
			return c;
		crc = updcrc(c, crc);
		*hdr = c;
	}
	if ((c = zdlread()) & ~0377)
		return c;
	crc = updcrc(c, crc);
	if ((c = zdlread()) & ~0377)
		return c;
	crc = updcrc(c, crc);
	if (crc & 0xFFFF) {
		verbose_code(zperr(badcrc);)
		return ERROR;
	}
	Zmodem = 1;
	return Rxtype;
}


/* Receive a binary style header (type and position) with 32 bit FCS */
static int zrbhd32(char *hdr)
{
	int c, n;
	unsigned long crc;

	if ((c = zdlread()) & ~0377)
		return c;
	Rxtype = c;
	crc = 0xFFFFFFFFL; crc = UPDC32(c, crc);
	verbose_code(vfile("zrbhd32 c=%X  crc=%lX", c, crc);)

	for (n=Rxhlen; --n >= 0; ++hdr) {
		if ((c = zdlread()) & ~0377)
			return c;
		crc = UPDC32(c, crc);
		*hdr = c;
		verbose_code(vfile("zrbhd32 c=%X  crc=%lX", c, crc);)
	}
	for (n=4; --n >= 0;) {
		if ((c = zdlread()) & ~0377)
			return c;
		crc = UPDC32(c, crc);
		verbose_code(vfile("zrbhd32 c=%X  crc=%lX", c, crc);)
	}
	if (crc != 0xDEBB20E3) {
		verbose_code(zperr(badcrc);)
		return ERROR;
	}
	Zmodem = 1;
	return Rxtype;
}


/*
 * Read a character from the modem line with timeout.
 *  Eat parity, XON and XOFF characters.
 */
static int noxrd7(void)
{
	int c;

	for (;;) {
		c = readline(Rxtimeout, __LINE__);
		verbose_code(if (Verbose > 8)
			vfile("noxrd7 line %d: c=%02x", __LINE__, c);)
		if (c < 0)
			return c;

		switch (c &= 0177) {
		case XON:
		case XOFF:
			verbose_code(if (Verbose > 8)
				vfile("noxrd7 c=XON|XOFF");)
			continue;
		default:
			verbose_code(if (Verbose > 8)
				vfile("noxrd7 c=default");)
			if (Zctlesc && !(c & 0140))
				continue;
		case '\r':
		case '\n':
		case ZDLE:
			verbose_code(if (Verbose > 8)
				vfile("noxrd7 returned c=%02x", c);)
			return c;
		}
	}
}

static int zgeth1(void)
{
	int c, n;

	if ((c = noxrd7()) < 0)
		return c;
	n = c - '0';
	if (n > 9)
		n -= ('a' - ':');
	if (n & ~0xF)
		return ERROR;
	if ((c = noxrd7()) < 0)
		return c;
	c -= '0';
	if (c > 9)
		c -= ('a' - ':');
	if (c & ~0xF)
		return ERROR;
	c += (n<<4);
	return c;
}

/* Decode two lower case hex digits into an 8 bit byte value */
static int zgethex(void)
{
	int c;

	c = zgeth1();
	verbose_code(if (Verbose>8)
		vfile("zgethex: %02X", c);)
	return c;
}

/* Receive a hex style header (type and position) */
static int zrhhdr(char *hdr)
{
	int c;
	unsigned short crc;
	int n;

	if ((c = zgethex()) < 0)
		return c;
	Rxtype = c;
	crc = updcrc(c, 0);

	for (n=Rxhlen; --n >= 0; ++hdr) {
		if ((c = zgethex()) < 0)
			return c;
		crc = updcrc(c, crc);
		*hdr = c;
	}
	if ((c = zgethex()) < 0)
		return c;
	crc = updcrc(c, crc);
	if ((c = zgethex()) < 0)
		return c;
	crc = updcrc(c, crc);
	if (crc & 0xFFFF) {
		verbose_code(zperr(badcrc);)
		return ERROR;
	}
	c = readline(2, __LINE__);
	if (c < 0)
		return c;
	c = readline(2, __LINE__);
	Zmodem = 1;
	if (c < 0)
		return c;
	return Rxtype;
}

/*
 * Read a ZMODEM header to hdr, either binary or hex.
 *  eflag controls local display of non zmodem characters:
 *      0:  no display
 *      1:  display printing characters only
 *      2:  display all non ZMODEM characters
 *
 *   Set Rxhlen to size of header (default 4) (valid iff good hdr)
 *  On success, set Zmodem to 1, set Rxpos and return type of header.
 *   Otherwise return negative on error.
 *   Return ERROR instantly if ZCRCW sequence, for fast error recovery.
 */
static int zgethdr(char *hdr, int eflag)
{
	int c, n, cancount;

	n = Zrwindow + Effbaud;         /* Max bytes before start of frame */
	Rxframeind = Rxtype = 0;

startover:
	cancount = 5;
again:
	/* Return immediate ERROR if ZCRCW sequence seen */
	switch (c = readline(Rxtimeout, __LINE__)) {
	case 021: case 0221:
		goto again;
	case RCDO:
	case TIMEOUT:
		goto fifi;
	case CAN:
gotcan:
		if (--cancount <= 0) {
			c = ZCAN; goto fifi;
		}
		switch (c = readline(1, __LINE__)) {
		case TIMEOUT:
			goto again;
		case ZCRCW:
			switch (readline(1, __LINE__)) {
			case TIMEOUT:
				c = ERROR; goto fifi;
			case RCDO:
				goto fifi;
			default:
				goto agn2;
			}
		case RCDO:
			goto fifi;
		default:
			break;
		case CAN:
			if (--cancount <= 0) {
				c = ZCAN; goto fifi;
			}
			goto again;
		}
	/* **** FALL THRU TO **** */
	default:
agn2:
		if ( --n == 0) {
			c = GCOUNT;  goto fifi;
		}
		goto startover;
	case ZPAD:              /* This is what we want. */
		break;
	}
	cancount = 5;
splat:
	switch (c = noxrd7()) {
	case ZPAD:
		goto splat;
	case RCDO:
	case TIMEOUT:
		goto fifi;
	default:
		goto agn2;
	case ZDLE:              /* This is what we want. */
		break;
	}

	Rxhlen = 4;             /* Set default length */
	Rxframeind = c = noxrd7();
	switch (c) {
	case ZVBIN32:
		if ((Rxhlen = c = zdlread()) < 0)
			goto fifi;
		if (c > ZMAXHLEN)
			goto agn2;
		Crc32r = 1;  c = zrbhd32(hdr); break;
	case ZBIN32:
		if (Usevhdrs)
			goto agn2;
		Crc32r = 1;  c = zrbhd32(hdr); break;
	case ZVBINR32:
		if ((Rxhlen = c = zdlread()) < 0)
			goto fifi;
		if (c > ZMAXHLEN)
			goto agn2;
		Crc32r = 2;  c = zrbhd32(hdr); break;
	case ZBINR32:
		if (Usevhdrs)
			goto agn2;
		Crc32r = 2;  c = zrbhd32(hdr); break;
	case RCDO:
	case TIMEOUT:
		goto fifi;
	case ZVBIN:
		if ((Rxhlen = c = zdlread()) < 0)
			goto fifi;
		if (c > ZMAXHLEN)
			goto agn2;
		Crc32r = 0;  c = zrbhdr(hdr); break;
	case ZBIN:
		if (Usevhdrs)
			goto agn2;
		Crc32r = 0;  c = zrbhdr(hdr); break;
	case ZVHEX:
		if ((Rxhlen = c = zgethex()) < 0)
			goto fifi;
		if (c > ZMAXHLEN)
			goto agn2;
		Crc32r = 0;  c = zrhhdr(hdr); break;
	case ZHEX:
		if (Usevhdrs)
			goto agn2;
		Crc32r = 0;  c = zrhhdr(hdr); break;
	case CAN:
		goto gotcan;
	default:
		goto agn2;
	}
	for (n = Rxhlen; ++n < ZMAXHLEN; )      /* Clear unused hdr bytes */
		hdr[n] = 0;
	Rxpos = hdr[ZP3] & 0377;
	Rxpos = (Rxpos<<8) + (hdr[ZP2] & 0377);
	Rxpos = (Rxpos<<8) + (hdr[ZP1] & 0377);
	Rxpos = (Rxpos<<8) + (hdr[ZP0] & 0377);
fifi:
	switch (c) {
	case GOTCAN:
		c = ZCAN;
	/* **** FALL THRU TO **** */
	case ZNAK:
	case ZCAN:
	case ERROR:
	case TIMEOUT:
	case RCDO:
	case GCOUNT:
		verbose_code(zperr("Got %s", frametypes[c+FTOFFSET]);)
	/* **** FALL THRU TO **** */
	default:
		verbose_code(if (c >= -4 && c <= FRTYPES)
			vfile("zgethdr: %c %d %s %lx", Rxframeind, Rxhlen,
			  frametypes[c+FTOFFSET], Rxpos);
		else
			vfile("zgethdr: %c %d %lx", Rxframeind, c, Rxpos);)
		;
	}
	/* Use variable length headers if we got one */
	if (c >= 0 && c <= FRTYPES && Rxframeind & 040) {
		Usevhdrs = 1;
		verbose_code(vfile("zgethdr: Usevhdrs = 1");)
		}
	return c;
}



/* Store long integer pos in Txhdr */
static void stohdr(unsigned long pos)
{
	Txhdr[ZP0] = pos;
	Txhdr[ZP1] = pos>>8;
	Txhdr[ZP2] = pos>>16;
	Txhdr[ZP3] = pos>>24;
}


/*
 * substr(string, token) searches for token in string s
 * returns pointer to token within string if found, NULL otherwise
 */
static const char * substr(const char *s, const char *t)
{
	const char *ss;
	const char *tt;
	/* search for first char of token */
	for (ss=s; *s; s++)
		if (*s == *t)
			/* compare token with substring */
			for (ss=s,tt=t; ;) {
				if (*tt == 0)
					return s;
				if (*ss++ != *tt++)
					break;
			}
	return NULL;
}

/* Say "bibi" to the receiver, try to do it cleanly */
static void saybibi(int line)
{

	verbose_code(vfile("saybibi: from %d line", line);)
	for (;;) {
		stohdr(0L);             /* CAF Was zsbhdr - minor change */
		zshhdr(4, ZFIN, Txhdr); /*  to make debugging easier */
		switch (zgethdr(Rxhdr, 0)) {
		case ZFIN:
			verbose_code(vfile("saybibi: ZFIN");)
			sendline('O'); sendline('O'); net_flush();
		case ZCAN:
		case TIMEOUT:
			return;
		}
	}
}

/* Send send-init information */
static int sendzsinit(void)
{
	int c;

	if (Myattn[0] == '\0' && (!Zctlesc || (Rxflags & TESCCTL)))
		return OK;
	errors = 0;
	for (;;) {
		stohdr(0L);
		if (Zctlesc) {
			Txhdr[ZF0] |= TESCCTL; zshhdr(4, ZSINIT, Txhdr);
		}
		else
			zsbhdr(4, ZSINIT, Txhdr);
		zsdata(Myattn, ZATTNLEN, ZCRCW);
		c = zgethdr(Rxhdr, 1);
		switch (c) {
		case ZCAN:
			return ERROR;
		case ZACK:
			return OK;
		default:
			if (++errors > 19)
				return ERROR;
			continue;
		}
	}
}

/*
 * Get the receiver's init parameters
 */
static int getzrxinit(void)
{
	int n;
	struct stat f;

	for (n=10; --n>=0; ) {

		switch (zgethdr(Rxhdr, 1)) {
		case ZCHALLENGE:        /* Echo receiver's challenge numbr */
			verbose_code(vfile("getzrxinit: ZCHALLENGE");)
			stohdr(Rxpos);
			zshhdr(4, ZACK, Txhdr);
			continue;
		case ZCOMMAND:          /* They didn't see out ZRQINIT */
			verbose_code(vfile("getzrxinit: ZCOMMAND");)
			stohdr(0L);
			zshhdr(4, ZRQINIT, Txhdr);
			continue;
		case ZRINIT:
			verbose_code(vfile("getzrxinit: ZRINIT");)
			Rxflags = 0377 & Rxhdr[ZF0];
			Usevhdrs = Rxhdr[ZF1] & CANVHDR;
			Txfcs32 = (Wantfcs32 && (Rxflags & CANFC32));
			Zctlesc |= Rxflags & TESCCTL;
			Rxbuflen = (0377 & Rxhdr[ZP0])+((0377 & Rxhdr[ZP1])<<8);
			if ( !(Rxflags & CANFDX))
				Txwindow = 0;
			verbose_code(vfile("Rxbuflen=%d Tframlen=%d", Rxbuflen, Tframlen);)
			signal(SIGINT, SIG_IGN);

			/* Override to force shorter frame length */
			if (Rxbuflen && (Rxbuflen>Tframlen) && (Tframlen>=32))
				Rxbuflen = Tframlen;
			if ( !Rxbuflen && (Tframlen>=32) && (Tframlen<=1024))
				Rxbuflen = Tframlen;
			verbose_code(vfile("Rxbuflen=%d", Rxbuflen);)


			/*
			 * If input is not a regular file, force ACK's to
			 *  prevent running beyond the buffer limits
			 */
			if ( !Command) {
				fstat(fileno(in), &f);
				if (!S_ISREG(f.st_mode)) {
					Canseek = -1;
					return ERROR;
				}
			}

			/* Set initial subpacket length */
			if (blklen < 1024) {    /* Command line override? */
				if (Effbaud > 300)
					blklen = 256;
				if (Effbaud > 1200)
					blklen = 512;
				if (Effbaud > 2400)
					blklen = 1024;
			}
			if (Rxbuflen && blklen>Rxbuflen)
				blklen = Rxbuflen;
			if (blkopt && blklen > blkopt)
				blklen = blkopt;
			verbose_code(vfile("Rxbuflen=%d blklen=%d", Rxbuflen, blklen);
			vfile("Txwindow=%u Txwspac=%d", Txwindow, Txwspac);)

			if (Lztrans == ZTRLE && (Rxflags & CANRLE))
				Txfcs32 = 2;
			else
				Lztrans = 0;

			return (sendzsinit());
		case ZCAN:
		case TIMEOUT:
			verbose_code(vfile("getzrxinit: ZCAN");)
			return ERROR;
		case ZRQINIT:
			if (Rxhdr[ZF0] == ZCOMMAND) {
				verbose_code(vfile("getzrxinit: ZRQINIT continue");)
				continue;
			} else {
				verbose_code(vfile("getzrxinit: ZRQINIT"))
				;
			}

		default:
			verbose_code(vfile("getzrxinit: default");)
			zshhdr(4, ZNAK, Txhdr);
			continue;
		}
	}
	verbose_code(vfile("getzrxinit: return ERROR");)
	return ERROR;
}


static int getnak(int line)
{
	int firstch;

	verbose_code(if (Verbose > 5)
	       vfile("Calling getnak from line %d", line);)

	Lastrx = 0;
	for (;;) {
		switch (firstch = readline(800, __LINE__)) {
		case ZPAD:
			verbose_code(if (Verbose > 8)
				vfile("readline from getnak returned ZPAD");)
			if (getzrxinit()) {
				verbose_code(if (Verbose > 8)
					vfile("getzrxinit from getnak returned!=0");)
				return ERROR;
			}
			return FALSE;
		case TIMEOUT:
			sprintf(endmsg, "Timeout waiting for ZRINIT");
			verbose_code(if (Verbose > 8)
				vfile("readline from getnak returned TIMEOUT");)
			return TRUE;
		case WANTG:
			verbose_code(if (Verbose > 8)
				vfile("readline from getnak returned WANTG");)
			Optiong = TRUE;
			blklen=1024;
		case WANTCRC:
			Crcflg = TRUE;
		case NAK:
			verbose_code(if (Verbose > 8)
				vfile(
				"readline from getnak returned %s",
				firstch==WANTG ? "WANTG" : firstch==WANTCRC ? "WANTCRC" : "NAK");)
			return FALSE;
		case CAN:
			if ((firstch = readline(20, __LINE__)) == CAN && Lastrx == CAN) {
				sprintf(endmsg, "Got CAN waiting to send file");
				return TRUE;
			}
		default:
			verbose_code(if (Verbose > 8)
				vfile("readline from getnak returned other");)
			break;
		}
		Lastrx = firstch;
	}
}

/* cseclen - data length of this sector to send */
static int wcputsec(const char *buf, int sectnum, int cseclen)
{
	int checksum, wcj;
	const char *cp;
	unsigned oldcrc;
	int firstch;
	int attempts;

	firstch=0;      /* part of logic to detect CAN CAN */

	verbose_code(if (Verbose>2)
		vfile("Sector %3d %2dk", Totsecs, Totsecs/8 );
	else if (Verbose>1)
		vfile("\rSector %3d %2dk", Totsecs, Totsecs/8 );)
	for (attempts=0; attempts <= RETRYMAX_SZ; attempts++) {
		Lastrx= firstch;
		sendline(cseclen==1024?STX:SOH);
		sendline(sectnum);
		sendline(-(sectnum + 1));
		oldcrc=checksum=0;
		for (wcj=cseclen,cp=buf; --wcj>=0; ) {
			sendline(*cp);
			oldcrc=updcrc((0377& *cp), oldcrc);
			checksum += *cp++;
		}
		if (Crcflg) {
			oldcrc=updcrc(0,updcrc(0,oldcrc));
			sendline((int)oldcrc>>8);
			sendline((int)oldcrc);
		}
		else
			sendline(checksum);
		net_flush();

		if (Optiong) {
			firstsec = FALSE; return OK;
		}
		firstch = readline(Rxtimeout, __LINE__);
gotnak:
		switch (firstch) {
		case CAN:
			if(Lastrx == CAN) {
cancan:
				verbose_code(zperr("Cancelled");)
				return ERROR;
			}
			break;
		case TIMEOUT:
			verbose_code(zperr("Timeout on sector ACK");)
			continue;
		case WANTCRC:
			if (firstsec)
				Crcflg = TRUE;
		case NAK:
			verbose_code(zperr("NAK on sector");)
			continue;
		case ACK:
			firstsec=FALSE;
			Totsecs += (cseclen>>7);
			return OK;
		case ERROR:
			verbose_code(zperr("Got burst for sector ACK");)
			break;
		default:
			verbose_code(zperr("Got %02x for sector ACK", firstch);)
			break;
		}
		for (;;) {
			Lastrx = firstch;
			if ((firstch = readline(Rxtimeout, __LINE__)) == TIMEOUT)
				break;
			if (firstch == NAK || firstch == WANTCRC)
				goto gotnak;
			if (firstch == CAN && Lastrx == CAN)
				goto cancan;
		}
	}
	verbose_code(zperr("Retry Count Exceeded");)
	return ERROR;
}

/* Fill buffer with blklen chars */
static int zfilbuf(void)
{
	int n;

	n = fread(txbuf, 1, blklen, in);
	if (n < blklen)
		Eofseen = 1;
	return n;
}

/*
 * Respond to receiver's complaint, get back in sync with receiver
 */
static int getinsync(int flag)
{
	int c=0;

	for (;;) {
		c = zgethdr(Rxhdr, 0);
		switch (c) {
		case ZCAN:
		case ZABORT:
		case ZFIN:
		case TIMEOUT:
			sprintf(endmsg, "Got %s sending data", frametypes[c+FTOFFSET]);
			return ERROR;
		case ZRPOS:
			if (Rxpos > bytcnt) {
				sprintf(endmsg, "Nonstandard Protocol");
				return ZRPOS;
			}
			/* ************************************* */
			/*  If sending to a buffered modem, you  */
			/*   might send a break at this point to */
			/*   dump the modem's buffer.            */
			clearerr(in);   /* In case file EOF seen */
			if (fseek(in, Rxpos, 0)) {
				sprintf(endmsg, "Bad Seek");
				return ERROR;
			}
			Eofseen = 0;
			bytcnt = Lrxpos = Txpos = Rxpos;
			if (Lastsync == Rxpos) {
				if (++Beenhereb4 > 12) {
					sprintf(endmsg, "Can't send block");
					return ERROR;
				}
				if (Beenhereb4 > 4)
					if (blklen > 32)
						blklen /= 2;
			}
			Lastsync = Rxpos;
			return c;
		case ZACK:
			Lrxpos = Rxpos;
			if (flag || Txpos == Rxpos)
				return ZACK;
			continue;
		case ZRINIT:
			return c;
		case ZSKIP:
			sprintf(endmsg, "File skipped by receiver request");
			return c;
		case ERROR:
		default:
			zsbhdr(4, ZNAK, Txhdr);
			continue;
		}
	}
}



/* Send the data in the file */
static int zsendfdata(void)
{
	int c, n;
	int newcnt;
	static int junkcount;  /* Counts garbage chars received by TX */
	/* static declaration special for not clobbered by longjmp */

	junkcount = 0;
	Beenhereb4 = FALSE;
somemore:
	if (setjmp(intrjmp)) {
waitack:
		junkcount = 0;
		c = getinsync(0);
gotack:
		switch (c) {
		default:
		case ZCAN:
			fclose(in);
			return ERROR;
		case ZRINIT:
			fclose(in);
			return ZSKIP;
		case ZSKIP:
			fclose(in);
			return c;
		case ZACK:
		case ZRPOS:
			break;
		}
		/*
		 * If the reverse channel can be tested for data,
		 *  this logic may be used to detect error packets
		 *  sent by the receiver, in place of setjmp/longjmp
		 *  net_rdchk() returns non 0 if a character is available
		 */
		while (net_rdchk()) {
			switch (readline(1, __LINE__))
			{
			case CAN:
			case ZPAD:
				c = getinsync(1);
				goto gotack;
			case XOFF:              /* Wait a while for an XON */
				readline(100, __LINE__);
			}
		}
	}

	signal(SIGINT, onintr);
	newcnt = Rxbuflen;
	Txwcnt = 0;
	stohdr(Txpos);
	zsbhdr(4, ZDATA, Txhdr);

	do {
		int e;

		n = zfilbuf();
		if (Eofseen)
			e = ZCRCE;
		else if (junkcount > 3)
			e = ZCRCW;
		else if (bytcnt == Lastsync)
			e = ZCRCW;
		else if (Rxbuflen && (newcnt -= n) <= 0)
			e = ZCRCW;
		else if (Txwindow && (Txwcnt += n) >= Txwspac) {
			Txwcnt = 0;  e = ZCRCQ;
		} else
			e = ZCRCG;
		verbose_code(if (Verbose>1)
			vfile("\r%7ld ZMODEM%s",
			  Txpos, Crc32t?" CRC-32":"");)
		zsdata(txbuf, n, e);
		bytcnt = Txpos += n;
		if (e == ZCRCW)
			goto waitack;
		/*
		 * If the reverse channel can be tested for data,
		 *  this logic may be used to detect error packets
		 *  sent by the receiver, in place of setjmp/longjmp
		 *  net_rdchk() returns non 0 if a character is available
		 */
		while (net_rdchk()) {
			switch (readline(1, __LINE__))
			{
			case CAN:
			case ZPAD:
				c = getinsync(1);
				if (c == ZACK)
					break;
				/* zcrce - dinna wanna starta ping-pong game */
				zsdata(txbuf, 0, ZCRCE);
				goto gotack;
			case XOFF:              /* Wait a while for an XON */
				readline(100, __LINE__);
			default:
				++junkcount;
			}
		}
		if (Txwindow) {
			long tcount;

			while ((tcount = (Txpos - Lrxpos)) >= Txwindow) {
				verbose_code(vfile("%ld window >= %u", tcount, Txwindow);)
				if (e != ZCRCQ)
					zsdata(txbuf, 0, e = ZCRCQ);
				c = getinsync(1);
				if (c != ZACK) {
					zsdata(txbuf, 0, ZCRCE);
					goto gotack;
				}
			}
			verbose_code(vfile("window = %ld", tcount);)
		}
	} while (!Eofseen);
	signal(SIGINT, SIG_IGN);

	if (Lrxpos != Txpos)
		n = Unlinkafter | Txwindow;
	else
		n = 0;
	for (;;) {
		stohdr(Txpos);
		if (n) {
			zsbhdr(4, ZDATA, Txhdr);
			zsdata(txbuf, 0, ZCRCW);
		} else {
			zsbhdr(4, ZEOF, Txhdr);
			}
		c = getinsync(0);
		switch (c) {
		case ZACK:
			n = 0;  continue;
		case ZNAK:
			continue;
		case ZRPOS:
			goto somemore;
		case ZRINIT:
			fclose(in);
			return OK;
		case ZSKIP:
			fclose(in);
			sprintf(endmsg, "File skipped by receiver request");
			return c;
		default:
			sprintf(endmsg, "Got %d trying to send end of file", c);
		case ERROR:
			fclose(in);
			return ERROR;
		}
	}
}

/* Send file name and related info */
static int zsendfile(const char *buf, int blen)
{
	int c;
	unsigned long crc = 0xFFFFFFFFL;
	long lastcrcrq = -1;

	verbose_code(if (Verbose > 8)
	       vfile("Calling zsendfile blen=%d", blen);)

	for (errors=0; ++errors<11;) {
		Txhdr[ZF0] = Lzconv;    /* file conversion request */
		Txhdr[ZF1] = Lzmanag;   /* file management request */
		if (Lskipnocor)
			Txhdr[ZF1] |= ZMSKNOLOC;
		Txhdr[ZF2] = Lztrans;   /* file transport request */
		Txhdr[ZF3] = 0;
		zsbhdr(4, ZFILE, Txhdr);
		zsdata(buf, blen, ZCRCW);
again:
		c = zgethdr(Rxhdr, 1);
		switch (c) {
		case ZRINIT:
			verbose_code(if (Verbose > 8)
				vfile("zgethdr from line=%d returned=ZRINIT", __LINE__-4);)
			while ((c = readline(50, __LINE__)) > 0)
				if (c == ZPAD) {
					goto again;
				}
			continue;
		case ZCAN:
		case TIMEOUT:
		case ZABORT:
		case ZFIN:
			verbose_code(if (Verbose > 8)
				vfile("zgethdr from line=%d returned=%s", __LINE__-14, frametypes[c+FTOFFSET]);)
			sprintf(endmsg, "Got %s on pathname", frametypes[c+FTOFFSET]);
			return ERROR;
		default:
			verbose_code(if (Verbose > 8)
				vfile("zgethdr from line=%d returned=other", __LINE__-19);)
			sprintf(endmsg, "Got %d frame type on pathname", c);
			continue;
		case ERROR:
		case ZNAK:
			verbose_code(if (Verbose > 8)
				vfile("zgethdr from line=%d returned=ERROR|ZNAK", __LINE__-25);)
			continue;
		case ZCRC:
			verbose_code(if (Verbose > 8)
				vfile("zgethdr from line=%d returned=ZCRC", __LINE__-29);)
			if (Rxpos != lastcrcrq) {
				lastcrcrq = Rxpos;
				crc = 0xFFFFFFFFL;
				if (Canseek >= 0) {
					fseek(in, 0L, 0);
					while (((c = getc(in)) != EOF) && --lastcrcrq)
						crc = UPDC32(c, crc);
					crc = ~crc;
					clearerr(in);   /* Clear possible EOF */
					lastcrcrq = Rxpos;
				}
			}
			stohdr(crc);
			zsbhdr(4, ZCRC, Txhdr);
			goto again;
		case ZFERR:
		case ZSKIP:
			verbose_code(if (Verbose > 8)
				vfile("zgethdr from line=%d returned=ZFERR|ZSKIP", __LINE__-48);)
			sprintf(endmsg, "File skipped by receiver request");
			fclose(in); return c;
		case ZRPOS:
			verbose_code(if (Verbose > 8)
				vfile("zgethdr from line=%d returned=ZRPOS", __LINE__-53);)
			/*
			 * Suppress zcrcw request otherwise triggered by
			 * lastyunc==bytcnt
			 */
			if (fseek(in, Rxpos, 0))
				return ERROR;
			Lastsync = (bytcnt = Txpos = Lrxpos = Rxpos) -1;
			return zsendfdata();
		}
	}
	verbose_code(if (Verbose > 8)
		vfile("zsendfile returned=ERROR");)
	fclose(in); return ERROR;
}

/*
 * generate and transmit pathname block consisting of
 *  pathname (null terminated),
 *  file length, mode time and file mode in octal
 *  as provided by the Unix fstat call.
 *  N.B.: modifies the passed name, may extend it!
 */
static int wctxpn(char *name)
{
	char *p, *q;
	char name2[PATHLEN];
	struct stat f;

	if (Modem2) {
		if (*name && fstat(fileno(in), &f)!= -1) {
			fprintf(stderr, "Sending %s, %ld XMODEM blocks. ",
			  name, (127+f.st_size)>>7);
		}
		fprintf(stderr, "Give your local XMODEM receive command now.\r\n");
		return OK;
	}
	verbose_code(zperr("Awaiting pathname nak for %s", *name?name:"<END>");)
	if ( !Zmodem)
		if (getnak(__LINE__)) {
			verbose_code(if (Verbose > 5)
				vfile("wctxpn: getnak returned!=0, returning ERROR");)
			return ERROR;
		}

	q = (char *) 0;
	if (Dottoslash) {               /* change . to / */
		for (p=name; *p; ++p) {
			if (*p == '/')
				q = p;
			else if (*p == '.')
				*(q=p) = '/';
		}
		if (q && strlen(++q) > 8) {     /* If name>8 chars */
			q += 8;                 /*   make it .ext */
			strcpy(name2, q);       /* save excess of name */
			*q = '.';
			strcpy(++q, name2);     /* add it back */
		}
	}

	for (p=name, q=txbuf ; *p; )
		if ((*q++ = *p++) == '/' && !Fullname)
			q = txbuf;
	*q++ = 0;
	p=q;
	while (q < (txbuf + 1024))
		*q++ = 0;
	if (*name) {
		if (fstat(fileno(in), &f)!= -1)
			sprintf(p, "%lu %lo %o 0 %d %ld", f.st_size, f.st_mtime,
			  f.st_mode, Filesleft, Totalleft);
		Totalleft -= f.st_size;
	}
	if (--Filesleft <= 0)
		Filesleft = Totalleft = 0;
	if (Totalleft < 0)
		Totalleft = 0;

	/* force 1k blocks if name won't fit in 128 byte block */
	if (txbuf[125])
		blklen=1024;
	else {          /* A little goodie for IMP/KMD */
		txbuf[127] = (f.st_size + 127) >>7;
		txbuf[126] = (f.st_size + 127) >>15;
	}
	if (Zmodem)
		return zsendfile(txbuf, 1+strlen(p)+(p-txbuf));
	if (wcputsec(txbuf, 0, 128)==ERROR)
		return ERROR;
	return OK;
}

/* fill buf with count chars padding with ^Z for CPM */
static int filbuf(char *buf, int count)
{
	int m;

	m = read(fileno(in), buf, count);
	if (m <= 0)
		return 0;
	while (m < count)
		buf[m++] = 032;
	return count;
}


static int wctx(unsigned long flen)
{
	int thisblklen;
	int sectnum, attempts, firstch;
	unsigned long charssent;

	charssent = 0;  firstsec=TRUE;  thisblklen = blklen;
	verbose_code(vfile("wctx:file length=%ld", flen);)

	while ((firstch=readline(Rxtimeout, __LINE__))!=NAK && firstch != WANTCRC
	  && firstch != WANTG && firstch!=TIMEOUT && firstch!=CAN)
		;
	if (firstch==CAN) {
		verbose_code(zperr("Receiver CANcelled");)
		return ERROR;
	}
	if (firstch==WANTCRC)
		Crcflg=TRUE;
	if (firstch==WANTG)
		Crcflg=TRUE;
	sectnum=0;
	for (;;) {
		if (flen <= (charssent + 896L))
			thisblklen = 128;
		if ( !filbuf(txbuf, thisblklen))
			break;
		if (wcputsec(txbuf, ++sectnum, thisblklen)==ERROR)
			return ERROR;
		charssent += thisblklen;
	}
	fclose(in);
	attempts=0;
	do {
		purgeline();
		sendline(EOT);
		net_flush();
		++attempts;
	}
		while ((firstch=(readline(Rxtimeout, __LINE__)) != ACK) && attempts < RETRYMAX_SZ);
	if (attempts == RETRYMAX_SZ) {
		verbose_code(zperr("No ACK on EOT");)
		return ERROR;
	}
	else
		return OK;
}

static int wcs(const char *oname)
{
	struct stat f;
	char name[PATHLEN];

	strcpy(name, oname);

	if (Restricted) {
		/* restrict pathnames to current tree or uucppublic */
		if ( substr(name, "../")
		 || (name[0]== '/' && strncmp(name, PUBDIR, strlen(PUBDIR))) ) {
			canit();  sprintf(endmsg,"Security Violation");
			return ERROR;
		}
	}

	in=fopen(oname, "r");

	if (in==NULL) {
		++errcnt;
		return OK;      /* pass over it, there may be others */
	}
	Eofseen = 0;  vpos = 0;

	/* Check for directory */
	fstat(fileno(in), &f);
	if (S_ISDIR(f.st_mode))
	{
		fclose(in);
		return OK;
	}

	++Filcnt;
	switch (wctxpn(name)) {
	case ZSKIP:
		if (Unlinkafter)
			unlink(oname);
	case ZFERR:
		return OK;
	case OK:
		break;
	default:
		return ERROR;
	}
	if (!Zmodem && wctx(f.st_size))
		return ERROR;

	if (Unlinkafter)
		unlink(oname);

	return 0;
}

/* Send command and related info */
static int zsendcmd(const char *buf, int blen)
{
	int c;
	long cmdnum;

	cmdnum = getpid();
	errors = 0;
	for (;;) {
		stohdr(cmdnum);
		Txhdr[ZF0] = Cmdack1;
		zsbhdr(4, ZCOMMAND, Txhdr);
		zsdata(buf, blen, ZCRCW);
listen:
		Rxtimeout = 100;                /* Ten second wait for resp. */
		Usevhdrs = 0;           /* Allow rx to send fixed len headers */
		c = zgethdr(Rxhdr, 1);

		switch (c) {
		case ZRINIT:
			goto listen;    /* CAF 8-21-87 */
		case ERROR:
		case GCOUNT:
		case TIMEOUT:
			if (++errors > Cmdtries)
				return ERROR;
			continue;
		case ZCAN:
		case ZABORT:
		case ZFIN:
		case ZSKIP:
		case ZRPOS:
			return ERROR;
		default:
			if (++errors > 20)
				return ERROR;
			continue;
		case ZCOMPL:
			Exitcode = Rxpos;
			saybibi(__LINE__);
			return OK;
		case ZRQINIT:
			verbose_code(vfile("******** RZ *******");)
			system("rz");
			verbose_code(vfile("******** SZ *******");)
			goto listen;
		}
	}
}

static void countem(int argc, char **argv)
{
	struct stat f;

	for (Totalleft = 0, Filesleft = 0; --argc >=0; ++argv) {
		f.st_size = -1;
		verbose_code(if (Verbose>2)
			vfile("\nCountem: %03d %s", argc, *argv);)

		if (access(*argv, 04) >= 0 && stat(*argv, &f) >= 0) {
			++Filesleft;  Totalleft += f.st_size;
		}
		verbose_code(if (Verbose>2)
			vfile( " %ld", f.st_size);)
	}
	verbose_code(if (Verbose>2)
		vfile("\ncountem: Total %d %ld",
		  Filesleft, Totalleft);)
}

static int wcsend(int argc, char *argp[])
{
	int n;

	Crcflg=FALSE;
	firstsec=TRUE;
	bytcnt = -1;
	if (Nozmodem) {
		printf("Start your YMODEM receive. ");  fflush(stdout);
	}
	for (n=0; n<argc; ++n) {
		Totsecs = 0;
		if (wcs(argp[n])==ERROR)
			return ERROR;
	}
	Totsecs = 0;
	if (Filcnt==0) {        /* bitch if we couldn't open ANY files */
		if (!Nozmodem && !Modem2) {
			Command = TRUE;
			Cmdstr = "echo \"sz: Can't open any requested files\"";
			if (getnak(__LINE__)) {
				Exitcode=1; canit();
			}
			if (!Zmodem)
				canit();
			else if (zsendcmd(Cmdstr, 1+strlen(Cmdstr))) {
				Exitcode=1; canit();
			}
			Exitcode = 1; return OK;
		}
		canit();
		sprintf(endmsg, "Can't open any requested files");
		return ERROR;
	}
	if (Zmodem)
		saybibi(__LINE__);
	else if ( !Modem2)
		wctxpn("");
	return OK;
}


#ifdef DIONIS
static
#endif
int sz(int argc, char *argv[])
{
	char *cp;
	int npats;
	char **patts = 0;

	atexit(atexit_fun);

	if ((cp = getenv("ZNULLS")) && *cp)
		Znulls = atoi(cp);
	if ((cp=getenv("SHELL")) && (substr(cp, "rsh") || substr(cp, "rksh")))
		Restricted=TRUE;

	chkinvok_sz(argv[0]);
	inittty(3);
	Rxtimeout = 600;
	npats=0;
	if (argc<2)
		usage_sz();
	while (--argc) {
		cp = *++argv;
		if (*cp++ == '-' && *cp) {
			while ( *cp) {
				switch(*cp++) {
				case '\\':
					 *cp = toupper(*cp);  continue;
				case '+':
					Lzmanag = ZMAPND; break;
#ifdef CSTOPB
				case '2':
					Twostop = TRUE; break;
#endif
				case 'a':
					if (Nozmodem || Modem2)
						usage_sz();
					Lzconv = ZCNL;  break;
				case 'b':
					Lzconv = ZCBIN; break;
				case 'C':
					if (--argc < 1) {
						usage_sz();
					}
					Cmdtries = atoi(*++argv);
					break;
				case 'c':
					Lzmanag = ZMCHNG;  break;
				case 'd':
					++Dottoslash;
					/* **** FALL THROUGH TO **** */
				case 'f':
					Fullname=TRUE; break;
				case 'g' :
					Ksendstr = TRUE; break;
				case 'e':
					Zctlesc = 1; break;
				case 'k':
					blklen=1024; break;
				case 'L':
					if (--argc < 1) {
						usage_sz();
					}
					blkopt = atoi(*++argv);
					if (blkopt<24 || blkopt>1024)
						usage_sz();
					break;
				case 'l':
					if (--argc < 1) {
						usage_sz();
					}
					Tframlen = atoi(*++argv);
					if (Tframlen<32 || Tframlen>1024)
						usage_sz();
					break;
				case 'N':
					Lzmanag = ZMNEWL;  break;
				case 'n':
					Lzmanag = ZMNEW;  break;
				case 'o':
					Wantfcs32 = FALSE; break;
				case 'p':
					Lzmanag = ZMPROT;  break;
				case 'r':
					if (Lzconv == ZCRESUM)
						Lzmanag = (Lzmanag & ZMMASK) | ZMCRC;
					Lzconv = ZCRESUM; break;
				case 't':
					if (--argc < 1) {
						usage_sz();
					}
					Rxtimeout = atoi(*++argv);
					if (Rxtimeout<10 || Rxtimeout>1000)
						usage_sz();
					break;
				case 'u':
					++Unlinkafter; break;
				case 'v':
					verbose_code(++Verbose;)
					break;
				case 'w':
					if (--argc < 1) {
						usage_sz();
					}
					Txwindow = atoi(*++argv);
					if (Txwindow < 256)
						Txwindow = 256;
					Txwindow = (Txwindow/64) * 64;
					Txwspac = Txwindow/4;
					if (blkopt > Txwspac
					 || (!blkopt && Txwspac < 1024))
						blkopt = Txwspac;
					break;
				case 'Y':
					Lskipnocor = TRUE;
					/* **** FALLL THROUGH TO **** */
				case 'y':
					Lzmanag = ZMCLOB; break;
				case 'Z':
				case 'z':
					Lztrans = ZTRLE;  break;
				default:
					usage_sz();
				}
			}
		}
		else if (Command) {
			if (argc != 1) {
				usage_sz();
			}
			Cmdstr = *argv;
		}
		else if ( !npats && argc>0) {
			if (argv[0][0]) {
				npats=argc;
				patts=argv;
			}
		}
	}
	if (npats < 1 && !Command)
		usage_sz();
	verbose_code(if (Verbose) {
		if (freopen(LOGFILE_SZ, "a", stderr)==NULL)
			if (freopen(LOGFILE2_SZ, "a", stderr)==NULL) {
				printf("Can't open log file!");
				exit(2);
			}
		setbuf(stderr, NULL);
	}
	vfile("%s %s for %s", Progname, VERSION, OS);)

	if (signal(SIGINT, bibi_sz) == SIG_IGN) {
		signal(SIGINT, SIG_IGN); signal(SIGKILL, SIG_IGN);
	} else {
		signal(SIGINT, bibi_sz); signal(SIGKILL, bibi_sz);
	}
#ifdef SIGQUIT
	signal(SIGQUIT, SIG_IGN);
#endif
#ifdef SIGTERM
	signal(SIGTERM, bibi_sz);
#endif

	countem(npats, patts);

	if (!Modem2 && !Nozmodem) {
		if (Ksendstr)
			printf(ksendbuf);
		printf("rz\r");  fflush(stdout);
		stohdr(0L);
		if (Command)
			Txhdr[ZF0] = ZCOMMAND;
		zshhdr(4, ZRQINIT, Txhdr);
	}
	net_flush();


	if (Command) {
		if (getzrxinit()) {
			Exitcode=1; canit();
		}
		else if (zsendcmd(Cmdstr, 1+strlen(Cmdstr))) {
			Exitcode=1; canit();
		}
	} else if (wcsend(npats, patts)==ERROR) {
		Exitcode=1;
		canit();
	}
	if (endmsg[0]) {
		printf("  %s: %s\r\n", Progname, endmsg);
		verbose_code(if (Verbose)
			vfile("%s\r", endmsg);)
	}
	printf("%s %s finished.\r\n", Progname, VERSION);
	fflush(stdout);
	if(errcnt || Exitcode)
		return(1);

	return(0);
}


/*
 *
 * rz.c By Chuck Forsberg
 */

#define LOGFILE_RZ "/tmp/rzlog"
#define LOGFILE2_RZ "rzlog"


/*
 * Routine to calculate the free bytes on the current file system
 *  ~0 means many free bytes (unknown)
 */
static inline unsigned long getfree()
{
	return(~0L);    /* many free bytes ... */
}

static unsigned long rxbytes;
static int Firstsec;

#define DEFBYTL 2000000000L     /* default rx file size */
static unsigned long Bytesleft; /* number of bytes of incoming file left */
static long Modtime;            /* Unix style mod time for incoming file */
static int  Filemode;           /* Unix style mode for incoming file */
static char Pathname[PATHLEN];

static int Batch;
static int Thisbinary;          /* current file is to be received in bin mode */
static int Rxbinary=FALSE;      /* receive all files in bin mode */
static int Rxascii=FALSE;       /* receive files in ascii (translate) mode */
static int Blklen;              /* record length of received packets */

#ifdef SEGMENTS
static int chinseg = 0; /* Number of characters received in this data seg */
static char secbuf[1+(SEGMENTS+1)*1024];
#else
static char secbuf[1025];
#endif

static struct utimbuf timep;
static char zconv;              /* ZMODEM file conversion request */
static char zmanag;             /* ZMODEM file management request */
static char ztrans;             /* ZMODEM file transport request */

/* Header type to send corresponding to Last rx close */
static int tryzhdrtype=ZRINIT;


/* called by signal interrupt or terminate to clean things up */
static void bibi_rz(int n)
{
	if (Zmodem)
		zmputs(Attn);
	canit();
	fprintf(stderr, "%s: caught signal %d; exiting", Progname, n);
	exit(3);
}

/*
 * If called as [-][dir/../]vrzCOMMAND set Verbose to 1
 * If called as rb use YMODEM protocol
 */
static void chkinvok_rz(const char *s)
{
	if (s[1]=='z')
		Batch = TRUE;
	else if (s[1]=='c')
		Crcflg = TRUE;
	else if (s[1]=='b')
		Batch = Nozmodem = TRUE;
}

static void usage_rz(void)
{
	inittty(0);
	fprintf(stderr,
"Receive Files and Commands with ZMODEM/YMODEM/XMODEM Protocol\n\n"
"%s %s for %s by Chuck Forsberg, Omen Technology INC\n"
"\t ported to telnet by Heo,Junhyeok, KAIST\n"
"\t\t\042The High Reliability Software\042\n"
"Usage:  rz [-v]         (ZMODEM)\n"
"or      rb [-av]        (YMODEM)\n"
"or      rc [-av] file   (XMODEM-CRC)\n"
"or      rx [-av] file   (XMODEM)\n\n"
"Supports incoming ZMODEM binary (-b), ASCII CR/LF>NL (-a), newer(-n),\n\
newer+longer(-N), protect (-p), Crash Recovery (-r),\n\
clobber (-y), match+clobber (-Y), compression (-Z), and append (-+).\n\n"
"Copyright 1993 Omen Technology INC All Rights Reserved\n"
"See rz.doc for option descriptions and licensing information.\n\n"
"Ported into telnet by Heo, Junhyeok, 1994.3.28\n"
"Linux porting by Park Myeong Seok,   1995.4.23\n"
	, Progname, VERSION, OS);
	exit(2);
}

#ifdef DIONIS
/* Hack! Special for Dionis emulation ONLY!
   NOT USE for real rz/sz!
*/
static int Dionis;

static void dionis_move_after_right_received(void)
{
	char *s;

	if(Dionis) {    /* Check full receive without errors? */
	    s = malloc(strlen(Pathname)+3); /* ..\0 */
	    if(s==NULL) {
		fprintf(stderr, "Not enought memory\n");
	    } else {
		if(rename(Pathname, strcat(strcpy(s, ".."), Pathname)))
			perror("rename()");
		free(s);
	    }
	}
}
#endif

/*
 * Close the receive dataset, return OK or ERROR
 */
static int closeit(void)
{
	time_t time();

	if (fclose(fout)==ERROR) {
		fprintf(stderr, "file close ERROR\n");
		return ERROR;
	}
	if (Modtime) {
		timep.actime = time(NULL);
		timep.modtime = Modtime;
		utime(Pathname, &timep);
	}
	if (S_ISREG(Filemode))
		chmod(Pathname, (unsigned short)(07777 & Filemode));
	return OK;
}


/*
 * Totalitarian Communist pathname processing
 */
static void checkpath(const char *name)
{
	if (Restricted) {
		if (fopen(name, "r") != NULL) {
			canit();
			fprintf(stderr, "\r\nrz: %s exists\n", name);
			bibi_rz(-1);
		}
		/* restrict pathnames to current tree or uucppublic */
		if ( substr(name, "../")
		 || (name[0]== '/' && strncmp(name, PUBDIR, strlen(PUBDIR))) ) {
			canit();
			fprintf(stderr,"\r\nrz:\tSecurity Violation\r\n");
			bibi_rz(-1);
		}
	}
}


static void openit(const char *name, const char *openmode)
{
	fout = fopen(name, openmode);
}

/*
 *  Directory-creating routines from Public Domain TAR by John Gilmore
 */

/*
 * After a file/link/symlink/dir creation has failed, see if
 * it's because some required directory was not present, and if
 * so, create all required dirs.
 */
static int make_dirs(char *pathname)
{
	char *p;                        /* Points into path */
	int madeone = 0;                /* Did we do anything yet? */
	int save_errno = errno;         /* Remember caller's errno */

	if (errno != ENOENT)
		return 0;               /* Not our problem */

	for (p = strchr(pathname, '/'); p != NULL; p = strchr(p+1, '/')) {
		/* Avoid mkdir of empty string, if leading or double '/' */
		if (p == pathname || p[-1] == '/')
			continue;
		/* Avoid mkdir where last part of path is '.' */
		if (p[-1] == '.' && (p == pathname+1 || p[-2] == '/'))
			continue;
		*p = 0;                         /* Truncate the path there */
		if ( !mkdir(pathname, 0777)) {  /* Try to create it as a dir */
			verbose_code(vfile("Made directory %s", pathname);)
			madeone++;              /* Remember if we made one */
			*p = '/';
			continue;
		}
		*p = '/';
		if (errno == EEXIST)            /* Directory already exists */
			continue;
		/*
		 * Some other error in the mkdir.  We return to the caller.
		 */
		break;
	}
	errno = save_errno;             /* Restore caller's errno */
	return madeone;                 /* Tell them to retry if we made one */
}


/*
 * Process incoming file information header
 */
static int procheader(char *name)
{
	char *p;
	const char *openmode;
	static int dummy;
	struct stat f;

	/* set default parameters and overrides */
	openmode = "w";
	Thisbinary = (!Rxascii) || Rxbinary;
	if (zconv == ZCBIN && Lzconv != ZCRESUM)
		Lzconv = zconv;                 /* Remote Binary override */
	if (Lzconv)
		zconv = Lzconv;
	if (Lzmanag)
		zmanag = Lzmanag;

	/*
	 *  Process ZMODEM remote file management requests
	 */
	if (!Rxbinary && zconv == ZCNL) /* Remote ASCII override */
		Thisbinary = 0;
	if (zconv == ZCBIN)     /* Remote Binary override */
		Thisbinary = TRUE;
	else if (zmanag == ZMAPND)
		openmode = "a";

	Bytesleft = DEFBYTL; Filemode = 0; Modtime = 0L;

	if (!name || !*name)
		return OK;

	p = name + 1 + strlen(name);
	if (*p) {       /* file coming from Unix or DOS system */
		sscanf(p, "%ld%lo%o%d%d%ld%d%d",
		  &Bytesleft, &Modtime, &Filemode,
		  &dummy, &Filesleft, &Totalleft, &dummy, &dummy);
		if (Filemode & UNIXFILE)
			++Thisbinary;
		verbose_code(if (Verbose) {
			vfile("Incoming: %s %ld %lo %o",
			  name, Bytesleft, Modtime, Filemode);
			vfile("YMODEM header: %s", p);
		})
	}


	else {          /* File coming from CP/M system */
		for (p=name; *p; ++p)           /* change / to _ */
			if ( *p == '/')
				*p = '_';

		if ( *--p == '.')               /* zap trailing period */
			*p = 0;
	}

	strcpy(Pathname, name);
	checkpath(name);

	if (*name && stat(name, &f)!= -1) {
		zmanag &= ZMMASK;
		verbose_code(vfile("Current %s is %ld %lo", name, f.st_size, f.st_mtime);)
		if (Thisbinary && zconv==ZCRESUM) {
			rxbytes = f.st_size & ~511;
			if (Bytesleft < rxbytes) {
				rxbytes = 0;  goto doopen;
			} else
				openit(name, "r+");
			if ( !fout)
				return ZFERR;
			if (fseek(fout, rxbytes, 0)) {
				closeit();
				return ZFERR;
			}
			verbose_code(vfile("Crash recovery at %ld", rxbytes);)
			return 0;
		}
		else if ((zmanag==ZMNEW) ||
		  ((zmanag==ZMNEWL) && Bytesleft <= f.st_size) ) {
			if ((f.st_mtime+1) >= Modtime)
				goto skipfile;
			goto doopen;
		}
		switch (zmanag & ZMMASK) {
		case ZMCLOB:
		case ZMAPND:
			goto doopen;
		default:
			goto skipfile;
		}
	} else if (zmanag & ZMSKNOLOC) {
skipfile:
		verbose_code(vfile("Skipping %s", name);)
		return ZSKIP;
	}
doopen:
	openit(name, openmode);

	if ( !fout)
		if (make_dirs(name))
			openit(name, openmode);

	if ( !fout)
		return ZFERR;
	return 0;
}

/*
 * Putsec writes the n characters of buf to receive file fout.
 *  If not in binary mode, carriage returns, and all characters
 *  starting with CPMEOF are discarded.
 */
static int putsec(char *buf, int n)
{
	char *p;

	if (n == 0)
		return OK;
	if (Thisbinary) {
		for (p=buf; --n>=0; )
			putc( *p++, fout);
	}
	else {
		if (Eofseen)
			return OK;
		for (p=buf; --n>=0; ++p ) {
			if ( *p == '\r')
				continue;
			if (*p == CPMEOF) {
				Eofseen=TRUE; return OK;
			}
			putc(*p ,fout);
		}
	}
	return OK;
}

/*
 * Receive a file with ZMODEM protocol
 *  Assumes file name frame is in secbuf
 */
static int rzfile(void)
{
	int c, n;

	Eofseen=FALSE;
	n = 20; rxbytes = 0l;

	if ((c = procheader(secbuf))!=0) {
		return (tryzhdrtype = c);
	}

	for (;;) {
#ifdef SEGMENTS
		chinseg = 0;
#endif
		stohdr(rxbytes);
		zshhdr(4,ZRPOS, Txhdr);
nxthdr:
		switch (c = zgethdr(Rxhdr, 0)) {
		default:
			verbose_code(vfile("rzfile: Wrong header %d", c);)
			if ( --n < 0) {
				sprintf(endmsg, "rzfile: Wrong header %d", c);
				return ERROR;
			}
			continue;
		case ZCAN:
			sprintf(endmsg, "Sender CANcelled");
			return ERROR;
		case ZNAK:
#ifdef SEGMENTS
			putsec(secbuf, chinseg);
			chinseg = 0;
#endif
			if ( --n < 0) {
				sprintf(endmsg, "rzfile: got ZNAK");
				return ERROR;
			}
			continue;
		case TIMEOUT:
#ifdef SEGMENTS
			putsec(secbuf, chinseg);
			chinseg = 0;
#endif
			if ( --n < 0) {
				sprintf(endmsg, "rzfile: TIMEOUT");
				return ERROR;
			}
			continue;
		case ZFILE:
			zrdata(secbuf, 1024);
			continue;
		case ZEOF:
#ifdef SEGMENTS
			putsec(secbuf, chinseg);
			chinseg = 0;
#endif
			if (rclhdr(Rxhdr) != rxbytes) {
				/*
				 * Ignore eof if it's at wrong place - force
				 *  a timeout because the eof might have gone
				 *  out before we sent our zrpos.
				 */
				errors = 0;  goto nxthdr;
			}
			if (closeit()) {
				tryzhdrtype = ZFERR;
				verbose_code(vfile("rzfile: closeit returned <> 0");)
				sprintf(endmsg,"Error closing file");
				return ERROR;
			}
			verbose_code(vfile("rzfile: normal EOF");)
#ifdef DIONIS
			dionis_move_after_right_received();
#endif
			return c;
		case ERROR:     /* Too much garbage in header search error */
#ifdef SEGMENTS
			putsec(secbuf, chinseg);
			chinseg = 0;
#endif
			if ( --n < 0) {
				sprintf(endmsg, "Persistent CRC or other ERROR");
				return ERROR;
			}
			zmputs(Attn);
			continue;
		case ZSKIP:
#ifdef SEGMENTS
			putsec(secbuf, chinseg);
			chinseg = 0;
#endif
			Modtime = 1;
			closeit();
			sprintf(endmsg, "Sender SKIPPED file");
			return c;
		case ZDATA:
			if (rclhdr(Rxhdr) != rxbytes) {
				if ( --n < 0) {
					sprintf(endmsg,"Data has bad addr");
					return ERROR;
				}
#ifdef SEGMENTS
				putsec(secbuf, chinseg);
				chinseg = 0;
#endif
				zmputs(Attn);  continue;
			}
moredata:
			verbose_code(if (Verbose>1)
				vfile( "\r%7ld ZMODEM%s",
				  rxbytes, Crc32r?" CRC-32":"");)
#ifdef SEGMENTS
			if (chinseg >= (1024 * SEGMENTS)) {
				putsec(secbuf, chinseg);
				chinseg = 0;
			}
			switch (c = zrdata(secbuf+chinseg, 1024))
#else
			switch (c = zrdata(secbuf, 1024))
#endif
			{
			case ZCAN:
#ifdef SEGMENTS
				putsec(secbuf, chinseg);
				chinseg = 0;
#endif
				sprintf(endmsg, "Sender CANcelled");
				return ERROR;
			case ERROR:     /* CRC error */
#ifdef SEGMENTS
				putsec(secbuf, chinseg);
				chinseg = 0;
#endif
				if ( --n < 0) {
					sprintf(endmsg, "Persistent CRC or other ERROR");
					return ERROR;
				}
				zmputs(Attn);
				continue;
			case TIMEOUT:
#ifdef SEGMENTS
				putsec(secbuf, chinseg);
				chinseg = 0;
#endif
				if ( --n < 0) {
					sprintf(endmsg, "TIMEOUT");
					return ERROR;
				}
				continue;
			case GOTCRCW:
				n = 20;
#ifdef SEGMENTS
				chinseg += Rxcount;
				putsec(secbuf, chinseg);
				chinseg = 0;
#else
				putsec(secbuf, Rxcount);
#endif
				rxbytes += Rxcount;
				stohdr(rxbytes);
				sendline(XON);
				zshhdr(4,ZACK, Txhdr);
				goto nxthdr;
			case GOTCRCQ:
				n = 20;
#ifdef SEGMENTS
				chinseg += Rxcount;
#else
				putsec(secbuf, Rxcount);
#endif
				rxbytes += Rxcount;
				stohdr(rxbytes);
				zshhdr(4,ZACK, Txhdr);
				goto moredata;
			case GOTCRCG:
				n = 20;
#ifdef SEGMENTS
				chinseg += Rxcount;
#else
				putsec(secbuf, Rxcount);
#endif
				rxbytes += Rxcount;
				goto moredata;
			case GOTCRCE:
				n = 20;
#ifdef SEGMENTS
				chinseg += Rxcount;
#else
				putsec(secbuf, Rxcount);
#endif
				rxbytes += Rxcount;
				goto nxthdr;
			}
		}
	}
}


/*
 * Strip leading ! if present, do shell escape.
 */
static int sys2(const char *s)
{
	if (*s == '!')
		++s;
	return system(s);
}
/*
 * Strip leading ! if present, do exec.
 */
static void exec2(const char *s)
{
	if (*s == '!')
		++s;
	execl("/bin/sh", "sh", "-c", s, 0);
}

/*
 * Ack a ZFIN packet, let byegones be byegones
 */
static void ackbibi(void)
{
	int n;

	verbose_code(vfile("ackbibi:");)
	Readnum = 1;
	stohdr(0L);
	for (n=3; --n>=0; ) {
		purgeline();
		zshhdr(4,ZFIN, Txhdr);
		switch (readline(100, __LINE__)) {
		case 'O':
			readline(1, __LINE__);    /* Discard 2nd 'O' */
			verbose_code(vfile("ackbibi complete");)
			return;
		case RCDO:
			return;
		case TIMEOUT:
		default:
			break;
		}
	}
}

/*
 * Initialize for Zmodem receive attempt, try to activate Zmodem sender
 *  Handles ZSINIT frame
 *  Return ZFILE if Zmodem filename received, -1 on error,
 *   ZCOMPL if transaction finished,  else 0
 */
static int tryz(void)
{
	int c, n;
	int cmdzack1flg;

	if (Nozmodem)           /* Check for "rb" program name */
		return 0;


	for (n=15; --n>=0; ) {
		/* Set buffer length (0) and capability flags */
#ifdef SEGMENTS
		stohdr(SEGMENTS*1024L);
#else
		stohdr(0L);
#endif
#ifdef CANBREAK
		Txhdr[ZF0] = CANFC32|CANFDX|CANOVIO|CANBRK;
#else
		Txhdr[ZF0] = CANFC32|CANFDX|CANOVIO;
#endif
		if (Zctlesc)
			Txhdr[ZF0] |= TESCCTL;
		Txhdr[ZF0] |= CANRLE;
		Txhdr[ZF1] = CANVHDR;
		/* tryzhdrtype may == ZRINIT */
		zshhdr(4,tryzhdrtype, Txhdr);
		if (tryzhdrtype == ZSKIP)       /* Don't skip too far */
			tryzhdrtype = ZRINIT;   /* CAF 8-21-87 */
again:
		switch (zgethdr(Rxhdr, 0)) {
		case ZRQINIT:
			if (Rxhdr[ZF3] & 0x80)
				Usevhdrs = 1;   /* we can var header */
			continue;
		case ZEOF:
			continue;
		case TIMEOUT:
			continue;
		case ZFILE:
			zconv = Rxhdr[ZF0];
			zmanag = Rxhdr[ZF1];
			ztrans = Rxhdr[ZF2];
			if (Rxhdr[ZF3] & ZCANVHDR)
				Usevhdrs = TRUE;
			tryzhdrtype = ZRINIT;
			c = zrdata(secbuf, 1024);
			if (c == GOTCRCW)
				return ZFILE;
			zshhdr(4,ZNAK, Txhdr);
			goto again;
		case ZSINIT:
			Zctlesc = TESCCTL & Rxhdr[ZF0];
			if (zrdata(Attn, ZATTNLEN) == GOTCRCW) {
				stohdr(1L);
				zshhdr(4,ZACK, Txhdr);
				goto again;
			}
			zshhdr(4, ZNAK, Txhdr);
			goto again;
		case ZFREECNT:
			stohdr(getfree());
			zshhdr(4, ZACK, Txhdr);
			goto again;
		case ZCOMMAND:
			cmdzack1flg = Rxhdr[ZF0];
			if (zrdata(secbuf, 1024) == GOTCRCW) {
				if (cmdzack1flg & ZCACK1)
					stohdr(0L);
				else
					stohdr((long)sys2(secbuf));
				purgeline();    /* dump impatient questions */
				do {
					zshhdr(4,ZCOMPL, Txhdr);
				}
				while (++errors<20 && zgethdr(Rxhdr,1) != ZFIN);
				ackbibi();
				if (cmdzack1flg & ZCACK1)
					exec2(secbuf);
				return ZCOMPL;
			}
			zshhdr(4, ZNAK, Txhdr); goto again;
		case ZCOMPL:
			goto again;
		default:
			continue;
		case ZFIN:
			ackbibi(); return ZCOMPL;
		case ZCAN:
			return ERROR;
		}
	}
	return 0;
}

/*
 * Receive 1 or more files with ZMODEM protocol
 */
static int rzfiles(void)
{
	int c;

	for (;;) {
		switch (c = rzfile()) {
		case ZEOF:
		case ZSKIP:
			switch (tryz()) {
			case ZCOMPL:
				return OK;
			default:
				return ERROR;
			case ZFILE:
				break;
			}
			continue;
		default:
			return c;
		case ERROR:
			return ERROR;
		}
	}
}


/*
 * Wcgetsec fetches a Ward Christensen type sector.
 * Returns sector number encountered or ERROR if valid sector not received,
 * or CAN CAN received
 * or WCEOT if eot sector
 * time is timeout for first char, set to 4 seconds thereafter
 ***************** NO ACK IS SENT IF SECTOR IS RECEIVED OK **************
 *    (Caller must do that when he is good and ready to get next sector)
 */

static int wcgetsec(char *rxbuf, int maxtime)
{
	int checksum, wcj, firstch;
	unsigned short oldcrc;
	char *p;
	int sectcurr;

	for (Lastrx=errors=0; errors<RETRYMAX_RZ; errors++) {

		if ((firstch=readline(maxtime, __LINE__))==STX) {
			Blklen=1024; goto get2;
		}
		if (firstch==SOH) {
			Blklen=128;
get2:
			sectcurr=readline(1, __LINE__);
			if ((sectcurr+(oldcrc=readline(1, __LINE__)))==0377) {
				oldcrc=checksum=0;
				for (p=rxbuf,wcj=Blklen; --wcj>=0; ) {
					if ((firstch=readline(1, __LINE__)) < 0)
						goto bilge;
					oldcrc=updcrc(firstch, oldcrc);
					checksum += (*p++ = firstch);
				}
				if ((firstch=readline(1, __LINE__)) < 0)
					goto bilge;
				if (Crcflg) {
					oldcrc=updcrc(firstch, oldcrc);
					if ((firstch=readline(1, __LINE__)) < 0)
						goto bilge;
					oldcrc=updcrc(firstch, oldcrc);
					if (oldcrc & 0xFFFF)
						verbose_code(zperr( "CRC"));
					else {
						Firstsec=FALSE;
						return sectcurr;
					}
				}
				else if (((checksum-firstch)&0377)==0) {
					Firstsec=FALSE;
					return sectcurr;
				}
				else
					verbose_code(zperr( "Checksum"));
			}
			else
				verbose_code(zperr("Sector number garbled"));
		}
		/* make sure eot really is eot and not just mixmash */
		else if (firstch==EOT && Lleft==0)
			return WCEOT;
		else if (firstch==CAN) {
			if (Lastrx==CAN) {
				verbose_code(zperr( "Sender CANcelled");)
				return ERROR;
			} else {
				Lastrx=CAN;
				continue;
			}
		}
		else if (firstch==TIMEOUT) {
			if (Firstsec)
				goto humbug;
bilge:
			verbose_code(zperr( "TIMEOUT"));
		}
		else
			verbose_code(zperr( "Got 0%o sector header", firstch));

humbug:
		Lastrx=0;
		while(readline(1, __LINE__)!=TIMEOUT)
			;
		if (Firstsec) {
			sendline(Crcflg?WANTCRC:NAK);  net_flush();
			Lleft=0;        /* Do read next time ... */
		} else {
			maxtime=40; sendline(NAK);  net_flush();
			Lleft=0;        /* Do read next time ... */
		}
	}
	/* try to stop the bubble machine. */
	canit();
	return ERROR;
}

/*
 * Fetch a pathname from the other end as a C ctyle ASCIZ string.
 * Length is indeterminate as long as less than Blklen
 * A null string represents no more files (YMODEM)
 */
static int wcrxpn(char *rpn) /* rpn - receive a pathname */
{
	int c;

	purgeline();

et_tu:
	Firstsec=TRUE;  Eofseen=FALSE;
	sendline(Crcflg?WANTCRC:NAK);  net_flush();
	Lleft=0;        /* Do read next time ... */
	while ((c = wcgetsec(rpn, 100)) != 0) {
		if (c == WCEOT) {
			verbose_code(zperr( "Pathname fetch returned %d", c);)
			sendline(ACK);  net_flush();
			Lleft=0;        /* Do read next time ... */
			readline(1, __LINE__);
			goto et_tu;
		}
		return ERROR;
	}
	sendline(ACK);  net_flush();
	return OK;
}


/*
 * Adapted from CMODEM13.C, written by
 * Jack M. Wierda and Roderick W. Hart
 */

static int wcrx(void)
{
	int sectnum, sectcurr;
	char sendchar;
	int cblklen;                    /* bytes to dump this block */

	Firstsec=TRUE;sectnum=0; Eofseen=FALSE;
	sendchar=Crcflg?WANTCRC:NAK;

	for (;;) {
		sendline(sendchar);     /* send it now, we're ready! */
		net_flush();
		Lleft=0;        /* Do read next time ... */
		sectcurr=wcgetsec(secbuf, (sectnum&0177)?50:130);
		if (sectcurr==((sectnum+1) &0377)) {
			sectnum++;
			cblklen = Bytesleft>Blklen ? Blklen:Bytesleft;
			if (putsec(secbuf, cblklen)==ERROR)
				return ERROR;
			if ((Bytesleft-=cblklen) < 0)
				Bytesleft = 0;
			sendchar=ACK;
		}
		else if (sectcurr==(sectnum&0377)) {
			verbose_code(zperr( "Received dup Sector");)
			sendchar=ACK;
		}
		else if (sectcurr==WCEOT) {
			if (closeit())
				return ERROR;
			sendline(ACK); net_flush();
			Lleft=0;        /* Do read next time ... */
#ifdef DIONIS
			dionis_move_after_right_received();
#endif
			return OK;
		}
		else if (sectcurr==ERROR)
			return ERROR;
		else {
			verbose_code(zperr( "Sync Error");)
			return ERROR;
		}
	}
}

/*
 * Let's receive something already.
 */
static int wcreceive(int argc, char **argp)
{
	int c;

	if (Batch || argc==0) {
		Crcflg=1;
		fprintf(stderr,
		   "%s ready. Type \"%s file ...\" to your modem program\n\r",
						Progname, Nozmodem?"sb":"sz");
		if ((c=tryz())!=0) {
			if (c == ZCOMPL)
				return OK;
			if (c == ERROR)
				goto fubar;
			c = rzfiles();
			if (c)
				goto fubar;
		} else {
			for (;;) {
				if (wcrxpn(secbuf)== ERROR)
					goto fubar;
				if (secbuf[0]==0)
					return OK;
				if (procheader(secbuf) == ERROR)
					goto fubar;
				if (wcrx()==ERROR)
					goto fubar;
			}
		}
	} else {
		Bytesleft = DEFBYTL; Filemode = 0; Modtime = 0L;

		procheader(""); strcpy(Pathname, *argp); checkpath(Pathname);
		fprintf(stderr, "\nrz: ready to receive %s\r\n", Pathname);
		if ((fout=fopen(Pathname, "w")) == NULL)
			return ERROR;
		if (wcrx()==ERROR)
			goto fubar;
	}
	return OK;
fubar:
	canit();
	Modtime = 1;
	if (fout)
		fclose(fout);
	if (Restricted) {
		unlink(Pathname);
		fprintf(stderr, "\r\nrz: %s removed.\r\n", Pathname);
	}
	return ERROR;
}


#ifdef DIONIS
static
#endif
int rz(int argc, char *argv[])
{
	char *cp;
	int npats;
	char *virgin, **patts = 0;
	int exitcode = 0;

	atexit(atexit_fun);

	Rxtimeout = 100;
	setbuf(stderr, NULL);
	if ((cp=getenv("SHELL")) && (substr(cp, "rsh") || substr(cp, "rksh")))
		Restricted=TRUE;

	chkinvok_rz(virgin=argv[0]);       /* if called as [-]rzCOMMAND set flag */
	inittty(3);
	npats = 0;
	while (--argc) {
		cp = *++argv;
		if (*cp == '-') {
			while( *++cp) {
				switch(*cp) {
				case '\\':
					 cp[1] = toupper(cp[1]);  continue;
				case 'a':
					if (!Batch || Nozmodem)
						Rxascii=TRUE;
					else
						usage_rz();
					break;
				case 't':
					if (--argc < 1) {
						usage_rz();
					}
					Rxtimeout = atoi(*++argv);
					if (Rxtimeout<10 || Rxtimeout>1000)
						usage_rz();
					break;
				case 'w':
					if (--argc < 1) {
						usage_rz();
					}
					Zrwindow = atoi(*++argv);
					break;
				case 'v':
					verbose_code(++Verbose;)
					break;
				case 'y':       /* dzo. Hmm. Forget? */
					break;
#ifdef DIONIS
				case 'D':
					Dionis++;
					break;
#endif
				default:
					usage_rz();
				}
			}
		}
		else if ( !npats && argc>0) {
			if (argv[0][0]) {
				npats=argc;
				patts=argv;
			}
		}
	}
	if (npats > 1)
		usage_rz();
	if (Batch && npats)
		usage_rz();
	verbose_code(if (Verbose) {

		if (freopen(LOGFILE_RZ, "a", stderr)==NULL)
			if (freopen(LOGFILE2_RZ, "a", stderr)==NULL) {
				printf("Can't open log file!");
				return(2);
			}
		setbuf(stderr, NULL);
		vfile("argv[0]=%s Progname=%s", virgin, Progname);
	})
	if (signal(SIGINT, bibi_rz) == SIG_IGN) {
		signal(SIGINT, SIG_IGN); signal(SIGKILL, SIG_IGN);
	}
	else {
		signal(SIGINT, bibi_rz); signal(SIGKILL, bibi_rz);
	}
	signal(SIGTERM, bibi_rz);
	if (wcreceive(npats, patts)==ERROR) {
		exitcode=1;
		canit();
	}
	if (exitcode && !Zmodem)        /* bellow again with all thy might. */
		canit();
	if (endmsg[0])
		printf("  %s: %s\r\n", Progname, endmsg);
	printf("%s %s finished.\r\n", Progname, VERSION);
	fflush(stdout);
	if(exitcode)
		return(exitcode);
	return(0);
}

#ifdef DIONIS

/* main function */

int main(int argc, char **argv)
{
	char *s = argv[0];
	char *applet_name;

	while(*s == '-')
		s++;
	for (applet_name = s; *s != '\0';) {
		if (*s++ == '/')
			applet_name = s;
	}
	while(*applet_name == '-')
		applet_name++;
	s = argv[0] = applet_name;
	if (*s == 'v') {
		verbose_code(Verbose=1;)
		++s;
	}
	Progname = s;
	if(*s == 'r' || *s == 'R')
		return(rz(argc, argv));
	return(sz(argc, argv));
}
#endif
